import numpy as np
import math

from scipy.spatial.transform import Rotation as R
from numpy.linalg import multi_dot, norm

def rotation_matrix(axis, theta):
    """
    Return the rotation matrix associated with counterclockwise rotation about
    the given axis by theta radians.
    
    example
    v = [3, 5, 0]
    axis = [4, 4, 1]
    theta = 1.2 
    
    np.dot(rotation_matrix(axis,theta), v)
    # [ 2.74911638  4.77180932  1.91629719]
    
    Ry = rotation_matrix([0,1,0], np.deg2rad(45))
    np.dot(Ry, [1,456,1]) 
    # [  1.41421356e+00   4.56000000e+02  -1.11022302e-16]

    """
    axis = np.asarray(axis)
    axis = axis/math.sqrt(np.dot(axis, axis))
    a = math.cos(theta/2.0)
    b, c, d = -axis*math.sin(theta/2.0)
    aa, bb, cc, dd = a*a, b*b, c*c, d*d
    bc, ad, ac, ab, bd, cd = b*c, a*d, a*c, a*b, b*d, c*d
    return np.array([[aa+bb-cc-dd, 2*(bc+ad), 2*(bd-ac)],
                     [2*(bc-ad), aa+cc-bb-dd, 2*(cd+ab)],
                     [2*(bd+ac), 2*(cd-ab), aa+dd-bb-cc]])

def rotate_points_around_arbitrary_axis(ps : np.ndarray, p1 : np.array, p2 : np.array, theta : float) -> np.ndarray:
    # see https://www.engr.uvic.ca/~mech410/lectures/4_2_RotateArbi.pdf
    """
    rotate_points_around_arbitrary_axis translates vectors (with points) to origin, do the rotation around axis and translate back to original positions

    :param np.ndarray ps: array with points
    :param np.array p1: first point describing axis
    :param np.array p2: second point describing axis
    :param float theta: angle in radians
    :return np.ndarray: rotated points
    """

    ps = np.vstack((ps.transpose(),np.array([1]*ps.transpose().shape[1])))

    x1, y1, z1 = p1
    axis = p2 - p1
  
    l = norm(axis)
    a,b,c = axis
    v = np.sqrt(b**2 + c**2)

    D = np.array([[1,0,0,-x1],[0,1,0,-y1],[0,0,1,-z1],[0,0,0,1]])
    D_inv = np.array([[1,0,0,x1],[0,1,0,y1],[0,0,1,z1],[0,0,0,1]])

    R_x = np.array([[1,0,0,0],[0,c/v,-b/v,0],[0,b/v,c/v,0],[0,0,0,1]])
    R_x_inv = np.array([[1,0,0,0],[0,c/v,b/v,0],[0,-b/v,c/v,0],[0,0,0,1]])

    R_y = np.array([[v/l,0,-a/l,0],[0,1,0,0],[a/l,0,v/l,0],[0,0,0,1]])
    R_y_inv = np.array([[v/l,0,a/l,0],[0,1,0,0],[-a/l,0,v/l,0],[0,0,0,1]])

    ct = math.cos(theta)
    st = math.sin(theta)
    R_z = np.array([[ct,-st,0,0],[st,ct,0,0],[0,0,1,0],[0,0,0,1]])

    ps_new = multi_dot([D_inv, R_x_inv, R_y_inv, R_z, R_y, R_x, D, ps])
    return ps_new[:3].transpose()


def rotate_points_around_origin_axis(ps : np.ndarray, axis : np.array, theta : float) -> np.ndarray:
    """
    rotate_points_around_origin_axis rotate points around axis (from origin)

    :param np.ndarray ps: array with points
    :param np.array axis: rotation axis
    :param float theta: angle in radians
    :return np.ndarray: rotated points
    """
    # origin is inside np.array([0, 0, 0])
    axis = axis / norm(axis)
    r = R.from_rotvec(theta * axis)
    return r.apply(ps)





import numpy as np
import pandas as pd
from pySailingVLM.inlet.winds import WindBase
from pySailingVLM.rotations.csys_transformations import CSYS_transformations

class InletConditions:
    def __init__(self, winds: WindBase, rho, center_of_pressure: np.ndarray):
        self.rho = rho  # fluid density [kg/m3]
        self.winds = winds
        self.cp_points = center_of_pressure
        self.tws_at_cp = np.array\
            ([self.winds.get_true_wind_speed_at_h(abs(cp_point[2])) for cp_point in self.cp_points])
        self.tws_length_at_ctr_points = np.linalg.norm(self.tws_at_cp, axis=1)

        self.V_app_infs = np.array\
            ([self.winds.get_app_infs_at_h(tws_at_ctr_point) for tws_at_ctr_point in self.tws_at_cp])
        self.V_app_infs_length = np.linalg.norm(self.V_app_infs, axis=1)

        self.AWA_infs_deg = np.rad2deg(np.arctan2(self.V_app_infs[:, 1], self.V_app_infs[:, 0]))

    def __iter__(self):
        yield 'cp_points_COG.x', self.cp_points[:, 0]
        yield 'cp_points_COG.y', self.cp_points[:, 1]
        yield 'cp_points_COG.z', self.cp_points[:, 2]
        yield 'TWS_COG.x', self.tws_at_cp[:, 0]
        yield 'TWS_COG.y', self.tws_at_cp[:, 1]
        yield 'TWS_COG.z', self.tws_at_cp[:, 2]
        yield 'TWS_length', self.tws_length_at_ctr_points
        yield 'V_app_infs_COG.x', self.V_app_infs[:, 0]
        yield 'V_app_infs_COG.y', self.V_app_infs[:, 1]
        yield 'V_app_infs_COG.z', self.V_app_infs[:, 2]
        yield 'V_app_infs_length', self.V_app_infs_length
        yield 'AWA_infs_COG_deg', self.AWA_infs_deg

    def to_df_full(self, csys_transformations: CSYS_transformations):
        obj_as_dict = dict(self)  # this calls __iter__

        tmp = {
            'AWA_infs_COW_deg': self.AWA_infs_deg - csys_transformations.leeway_deg,  # COW = centerline
        }
        obj_as_dict.update(tmp)
        df = pd.DataFrame.from_records(obj_as_dict)
        return df


import sys
import numpy as np
import matplotlib.pyplot as plt
import math 
class VlmAirfoil:
    # http://airfoiltools.com/airfoil/naca4digit
    # 1128 -> 0.01, 0.1, 0.28
    def __init__(self, max_camber : float, max_camber_position: float, thickness : float, n : int = 100):
        self._m = max_camber
        self._p = max_camber_position
        self._t = thickness
        self.n = n
        
        self.__a0 = 0.2969
        self.__a1 = -0.126
        self.__a2 = -0.3516
        self.__a3 = 0.2843
        self.__a4 = -0.1015
        self.__a4_te = -0.1036 # for x=1 (trailing edge)
        
        self.__check_naca_params()
        self._xc = np.linspace(0, 1, self.n)
        self._yc = self.__calculate_y_camber()
        
        self._yt = np.array([self.__get_camber_thickness(x) for x in self._xc])
        self.__dyc_dx = np.array([self.__get_dy_dx_camber(x) for x in self._xc])
        self.__theta = np.array([math.atan(x) for x in self.__dyc_dx])
        
        # u - upper, l - lower
        self._xu = self._xc - self._yt * np.sin(self.__theta)
        self._xl = self._xc + self._yt * np.sin(self.__theta)
        
        self._yu = self._yc + self._yt * np.cos(self.__theta)
        self._yl = self._yc - self._yt * np.cos(self.__theta)
    
    @property
    def xc(self) -> list:
        return self._xc
        
    @property
    def yc(self) -> list:
        return self._yc
    
    @property
    def yt(self) -> list:
        return self._yt
    
    @property
    def yu(self) -> list:
        return self._yu
    
    @property
    def yl(self) -> list:
        return self._yl
    
    @property
    def xu(self) -> list:
        return self._xu
    
    @property
    def xl(self) -> list:
        return self._xl
    
    def __check_naca_params(self):
        try:
            if not(self._m <= 0.3 and self._m >= 0):
                raise ValueError('Max camber must be between 0 and 30%!')
            
            if not(self._p <= 0.9 and self._p >= 0):
                raise ValueError('Max camber position must be between 0 and 90%!')
            
            if not(self._t <= 0.4 and self._t >= 0):
                raise ValueError('Thickness must be between 0 and 40%!')
        except ValueError as err:
            print(err.args[0])
            sys.exit(1)
    
    def __get_y_camber(self, x):
        # 0 <= x < p
        if x < self._p and x >= 0:
            return self._m * (2 * self._p * x - x**2) / (self._p**2)
        # p <= x <= 1
        elif x >= self._p and x <= 1:
            return self._m * (1 - 2*self._p + 2*self._p*x -x**2) / (1 - self._p)**2
        else:
            raise ValueError('x is not betwwen 0 and 1')
        
    def __get_dy_dx_camber(self, x):
        # 0 <= x < p
        if x < self._p and x >= 0:
            return 2 * self._m * (self._p  - x) / (self._p**2)
        # p <= x <= 1
        elif x >= self._p and x <= 1:
            return 2 * self._m * (self._p - x) / (1 - self._p)**2
        else:
            raise ValueError('x is not betwwen 0 and 1')
        
    def __get_camber_thickness(self, x):
        epsilon = 0.0001
        if 1.0 - x < epsilon:
            return 5 * self._t * (self.__a0 * math.sqrt(x) + self.__a1 *x + self.__a2 * x *x + self.__a3 * x * x * x + self.__a4_te * x * x * x * x) 
        else:
            return 5 * self._t * (self.__a0 * math.sqrt(x) + self.__a1 *x + self.__a2 * x *x + self.__a3 * x * x * x + self.__a4 * x * x * x * x) 
    
    def __calculate_y_camber(self):
        return np.array([self.__get_y_camber(x) for x in self._xc])
    
    def plot(self, show : bool = False):
        
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.set_xlim([0, 1])
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.axis('equal')
        ax.grid()

        ax.plot(self._xc, self._yc, '-', color='blue')

        plt.subplots_adjust(left=0.10, bottom=0.10, right=0.98, top=0.98, wspace=None, hspace=None)
        if show:
            plt.show()

   
      
### tests
# 9500
#foil = VlmAirfoil(-0.09, 0.5, 0.0, 100)
#foil.plot(True)



import numpy as np

from typing import List
from pySailingVLM.thin_airfoil.vlm_airfoil import VlmAirfoil
def discrete_segment(p1, p2, n):
    segment = []
    step = (p2 - p1) / n

    for i in range(n):
        point = p1 + i * step
        segment.append(point)

    segment.append(p2)
    return np.array(segment)


def make_point_mesh(segment1, segment2, n):
    mesh = []
    # TODO: dodac camber w stylu NACA airfoil --> wzorek z wikipedii
    # TODO: upewnic sie ze normalna jest w dobra strone dla pochylonej lodki
    for p1, p2 in zip(segment1, segment2):
        s = discrete_segment(p1, p2, n)
        mesh.append(np.array(s))

    return np.array(mesh)


def make_airfoil_mesh(le_points : List[np.ndarray], grid_size : List[int], chords_vec : np.ndarray, interpolated_distance_from_LE, interpolated_camber) -> np.ndarray:
    """
    make_airfoil_mesh make airfoil mesh before any rotation/twist etc

    :param List[np.ndarray] le_points: list with points on leading edge like [le_SW,  le_NW]
    :param List[int] grid_size: size of grid like [n_chordwise, n_spanwise]
    :param np.ndarray chords_vec: vector with chords
    :param _type_ interpolated_distance_from_LE: interpolated distance from leading edge
    :param _type_ interpolated_camber: interpolated camber
    :return np.ndarray: array with generated mesh
    """
    le_SW,  le_NW = le_points
    n_chordwise, n_spanwise = grid_size
    le_line = discrete_segment(le_SW, le_NW, n_spanwise)
    te_line = np.copy(le_line)  # deep copy
    te_line += chords_vec
    
    n = n_chordwise + 1
    segment1 = le_line
    segment2 = te_line
    distance = interpolated_distance_from_LE
    camber = interpolated_camber
    
    # segment1 , segemnt2 
    # http://www.airfoiltools.com/airfoil/naca4digit?MNaca4DigitForm%5Bcamber%5D=9&MNaca4DigitForm%5Bposition%5D=50&MNaca4DigitForm%5Bthick%5D=1&MNaca4DigitForm%5BnumPoints%5D=100&MNaca4DigitForm%5BcosSpace%5D=0&MNaca4DigitForm%5BcosSpace%5D=1&MNaca4DigitForm%5BcloseTe%5D=0&yt0=Plot
    # p is the position of the maximum camber divided by 10. In the example P=4 so the maximum camber is at 0.4 or 40% of the chord.
    # m is the maximum camber divided by 100. In the example M=2 so the camber is 0.02 or 2% of the chord
    mesh = []
    counter = 0
    for p1, p2 in zip(segment1, segment2):
        p = distance[counter]
        m = camber[counter] 
        foil = VlmAirfoil(m, p, 0.0, n=n)
        
        xs = (p2[0] - p1[0]) * foil.xc + p1[0]
        ys = (p2[0] - p1[0]) * foil.yc 
        zs =[p1[2]] * n 
    
        counter += 1
        mesh.append(list(zip(xs,ys, zs)))

    return np.array(mesh)
   
    


import os
import numpy as np
import time

half_wing_span = 0.5
chord_length = 1.0

mgirths = np.array([0.00, 1./8, 1./4, 1./2, 3./4, 7./8, 1.00])
mchords = np.array([chord_length]* len(mgirths))
mcenterline = -10+ 0* mgirths

jgirths = np.array([0.00, 1./4, 1./2, 3./4, 1.00])
jcenterline = 0*(10+5)  + 0*15. * jgirths

output_args = {
    'case_name': os.path.basename(__file__),  # get name of the current file
    'case_dir': os.path.dirname(os.path.realpath(__file__)), # get dir of the current file
    'name': os.path.join("results_example_jib_and_mainsail_vlm", time.strftime("%Y-%m-%d_%Hh%Mm%Ss")),
    'file_name': 'my_fancy_results', # name of xlsx excel file
}

# SOLVER SETTINGS
solver_args = {
    'n_spanwise':  16,  # No of control points (above the water) per sail, recommended: 50
    'n_chordwise': 8, # No of control points (above the water) per sail, recommended: 50
    'interpolation_type': "spline",  # either "spline" or "linear"
    'LLT_twist': "real_twist",  # defines how the Lifting Line discretize the sail twist.
}

# SAILING CONDITIONS
conditions_args = {
    'leeway_deg': 0.,    # [deg]
    'heel_deg': 0.,     # [deg]
    'SOG_yacht': 0.,   # [m/s] yacht speed - speed over ground (leeway is a separate variable)
    'tws_ref': 1.0,     # [m/s] true wind speed
    'alpha_true_wind_deg': 1.,   # [deg] true wind angle (with reference to course over ground) => Course Wind Angle to the boat track = true wind angle to centerline + Leeway
    'reference_water_level_for_wind_profile': -0.,  # [m] this is an attempt to mimick the deck effect
    # by lowering the sheer_above_waterline
    # while keeping the wind profile as in original geometry
    # this shall be negative (H = sail_ctrl_point - water_level)
    'wind_exp_coeff': 0.,  # [-] coefficient to determine the exponential wind profile
    'wind_reference_measurment_height': 10.,  # [m] reference height for exponential wind profile
    'rho': 1.225,  # air density [kg/m3]
    'wind_profile': 'flat', # allowed: 'exponential' or 'flat' or 'logarithmic'
    'roughness': 0.05, # for logarithmic profile only 
}


# GEOMETRY OF THE RIG
rig_args = {
    'main_sail_luff': 5.,  # [m]
    'jib_luff': 10.0,  # [m]
    'foretriangle_height': 11.50,  # [m]
    'foretriangle_base': 3.90,  # [m]
    'sheer_above_waterline': 0.,#[m]
    'boom_above_sheer': 0., # [m],
    'rake_deg': 90.,  # rake angle [deg]
    'mast_LOA': 0.,  # [m]
    'sails_def': 'main', # definition of sail set, possible: 'jib' or 'main' or 'jib_and_main'
}


# INPUT - GEOMETRY OF THE SAIL
# INFO for camber:
# First digit describing maximum camber as percentage of the chord.
# Second digit describing the distance of maximum camber from the airfoil leading edge in tenths of the chord.
main_sail_args = {
    'girths' : mgirths,
    'chords': mchords,
    'centerline_twist_deg': mcenterline,
    'camber': 0*np.array([0.01, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01]),
    'camber_distance_from_luff': np.array([0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5]),
}


jib_sail_args = {
    'centerline_twist_deg': jcenterline,
    'girths': jgirths,
    'chords': np.array([3.80, 2.98, 2.15, 1.33, 0.5]),
    'camber': 0*np.array([0.01, 0.01, 0.01, 0.01, 0.01]),
    'camber_distance_from_luff': np.array([0.5, 0.5, 0.5, 0.5, 0.5]), # starting from leading edge   
}        


# REFERENCE CSYS
# The origin of the default CSYS is located @ waterline level and aft face of the mast
# The positive x-coord: towards stern
# The positive y-coord: towards leeward side
# The positive z-coord: above the water
# To shift the default CSYS, adjust the 'reference_level_for_moments' variable.
# Shifted CSYS = original + reference_level_for_moments
# As a results the moments will be calculated around the new origin.

# yaw_reference [m] - distance from the aft of the mast towards stern, at which the yawing moment is calculated.
# sway_reference [m] - distance from the aft of the mast towards leeward side. 0 for symmetric yachts ;)
# heeling_reference [m] - distance from the water level,  at which the heeling moment is calculated.
csys_args = {
    'reference_level_for_moments': np.array([0, 0, 0]),  # [yaw_reference, sway_reference, heeling_reference]
}

# GEOMETRY OF THE KEEL
# to estimate heeling moment from keel, does not influence the optimizer.
# reminder: the z coord shall be negative (under the water)
keel_args={
    'center_of_lateral_resistance_upright': np.array([0, 0, -1.0]),  # [m] the coordinates for a yacht standing in upright position
} 


import numpy as np

from scipy.spatial.transform import Rotation as R

from pySailingVLM.solver.additional_functions import extract_above_water_quantities
from pySailingVLM.rotations.geometry_calc import rotation_matrix

class CSYS_transformations:
    def __init__(self, heel_deg=0, leeway_deg=0, v_from_original_xyz_2_reference_csys_xyz=np.array([0, 0, 0])):
        """
        v_from_original_2_reference_csys: np.array([reference_level_for_yaw_moment, 0, reference_level_for_heeling_moment])
        """
        self.heeling_rotation_matrix = rotation_matrix([1, 0, 0], np.deg2rad(-heel_deg))  # first rotation
        self.reverse_heeling_rotation_matrix = rotation_matrix([1, 0, 0], np.deg2rad(heel_deg))  # first rotation - for the mirror part
        self.leeway_rotation_matrix = rotation_matrix([0, 0, 1], np.deg2rad(leeway_deg))  # second rotation
        self.reverse_leeway_rotation_matrix = rotation_matrix([0, 0, 1], np.deg2rad(-leeway_deg))
        
        axis = [1, 0, 0]
        axis = axis / np.linalg.norm(axis)
        r = R.from_rotvec(np.deg2rad(-heel_deg) * axis)
        np.testing.assert_almost_equal(self.heeling_rotation_matrix, r.as_matrix())
        
        axis = [1, 0, 0]
        axis = axis / np.linalg.norm(axis)
        r = R.from_rotvec(np.deg2rad(heel_deg) * axis)
        np.testing.assert_almost_equal(self.reverse_heeling_rotation_matrix, r.as_matrix())
        
        axis = [0, 0, 1]
        axis = axis / np.linalg.norm(axis)
        r = R.from_rotvec(np.deg2rad(leeway_deg) * axis)
        np.testing.assert_almost_equal(self.leeway_rotation_matrix, r.as_matrix())
        
        axis = [0, 0, 1]
        axis = axis / np.linalg.norm(axis)
        r = R.from_rotvec(np.deg2rad(-leeway_deg) * axis)
        np.testing.assert_almost_equal(self.reverse_leeway_rotation_matrix, r.as_matrix())
        
        self.__leeway_deg = leeway_deg
        self.__heel_deg = heel_deg

        self.v_from_original_xyz_2_reference_csys_xyz = v_from_original_xyz_2_reference_csys_xyz

    def rotate_point_with_mirror(self, point):
        """
        from upright csys to heeled csys
        """
        if point[2] > 0:
            point = np.dot(self.heeling_rotation_matrix, point)
        else:
            point = np.dot(self.reverse_heeling_rotation_matrix, point)

        point = np.dot(self.leeway_rotation_matrix, point)
        return point

    def reverse_rotations_with_mirror(self, point):
        """
        from heeled csys to upright csys
        """
        point = np.dot(self.reverse_leeway_rotation_matrix, point)
        if point[2] > 0:
            point = np.dot(self.reverse_heeling_rotation_matrix, point)
        else:
            point = np.dot(self.heeling_rotation_matrix, point)

        return point

    def rotate_point_without_mirror(self, point):
        """
        from upright csys to heeled csys
        """
        point = np.dot(self.heeling_rotation_matrix, point)
        point = np.dot(self.leeway_rotation_matrix, point)
        return point

    def from_xyz_to_centerline_csys(self, array_of_xyz_vectors):
        """
        the x, y axes are parallel to water level
        the x axis is along the center line of the yacht, positive towards the stern
        the z axis is perpendicular to water level, positive upwards
        """
        array_of_vectors_in_centerline_csys = np.array([np.dot(self.leeway_rotation_matrix, v_xyz) for v_xyz in array_of_xyz_vectors])
        return array_of_vectors_in_centerline_csys

    def calc_centerline_moments(self, Mxyz):
        # side moment is around x axis
        # under water (mirror) moments are positive (righting)
        # above water moments are negative (heeling)
        M_centerline_csys = self.from_xyz_to_centerline_csys(Mxyz)
        _, total_above_water_moments_in_centerline_csys = extract_above_water_quantities(M_centerline_csys)
        return M_centerline_csys, total_above_water_moments_in_centerline_csys

    @property
    def leeway_deg(self):
        return self.__leeway_deg

    @property
    def heel_deg(self):
        return self.__heel_deg





from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pySailingVLM.runner.sail import Wind
    
import numpy as np
from dataclasses import dataclass, field

from pySailingVLM.inlet.inlet_conditions import InletConditions
from pySailingVLM.solver import coefs
from pySailingVLM.solver import forces
from pySailingVLM.solver import velocity
from pySailingVLM.solver import panels


# sprawdzic typy !!!!
@dataclass
class Vlm:

    panels: np.ndarray
    n_chordwise : int
    n_spanwise :int

    rho : float
    wind : np.ndarray
    trailing_edge_info : np.ndarray
    leading_edge_info : np.ndarray

    # post init atributes
    areas : np.ndarray = field(init=False)
    normals : np.ndarray = field(init=False)
    leading_mid_points : np.ndarray = field(init=False)
    trailing_mid_points : np.ndarray = field(init=False)
    ctr_p : np.ndarray = field(init=False) # control point in 3/4 of chord -> other name: collocation point 
    cp : np.ndarray = field(init=False) # 1/4 length of chord -> other name: center of pressure
    rings : np.ndarray = field(init=False)
    span_vectors : np.ndarray = field(init=False)
    coeff : np.ndarray = field(init=False)
    RHS : np.ndarray = field(init=False)
    v_ind_coeff : np.ndarray = field(init=False)
    gamma_magnitude : np.ndarray = field(init=False)
    inlet_conditions: InletConditions = field(init=False)
    force : np.ndarray = field(init=False)
    pressure : np.ndarray = field(init=False)
    V_app_fs_at_cp : np.ndarray = field(init=False)
    V_induced_at_cp : np.ndarray = field(init=False)
    p_coeffs: np.ndarray = field(init=False)
    # as class var
    gamma_orientation : float = -1.0

    def __post_init__(self):

        # M = wzdluz rozpoetosci skrzydel, spanwise
        # N = chordwise, linia laczaca leading i trailing
        self.areas = panels.get_panels_area(self.panels)
        #self.normals, self.ctr_p, self.cp, self.rings, self.span_vectors, self.leading_mid_points, self.trailing_mid_points = coefs.calculate_normals_collocations_cps_rings_spans_leading_trailing_mid_points(self.panels, self.gamma_orientation)
        self.normals, self.ctr_p, self.cp, self.rings, self.span_vectors, self.leading_mid_points, self.trailing_mid_points = coefs.calculate_vlm_variables(self.panels, self.trailing_edge_info, self.gamma_orientation, self.n_chordwise, self.n_spanwise)
        self.inlet_conditions = InletConditions(self.wind, self.rho, self.cp)
        self.coeff, self.v_ind_coeff = coefs.calc_velocity_coefs(self.inlet_conditions.V_app_infs, self.ctr_p, self.rings, self.normals, self.trailing_edge_info, self.gamma_orientation)
        self.RHS = coefs.calculate_RHS(self.inlet_conditions.V_app_infs, self.normals)
        self.gamma_magnitude = coefs.solve_eq(self.coeff, self.RHS)
        self.V_induced_at_ctrl,  self.V_app_fs_at_ctrl_p = velocity.calculate_app_fs(self.inlet_conditions.V_app_infs,  self.v_ind_coeff,  self.gamma_magnitude)

        # boundary condition calculated in collocation points (control points)
        assert forces.is_no_flux_BC_satisfied(self.V_app_fs_at_ctrl_p, self.panels, self.areas, self.normals)
        self.force, self.V_app_fs_at_cp, self.V_induced_at_cp = forces.calc_force_wrapper(self.inlet_conditions.V_app_infs, self.gamma_magnitude, self.rho, self.cp, self.rings, self.n_spanwise, self.normals, self.span_vectors, self.trailing_edge_info, self.leading_edge_info, self.gamma_orientation)
        self.pressure = forces.calc_pressure(self.force, self.normals, self.areas)

        self.p_coeffs = forces.calc_pressure_coeff(self.pressure, self.rho, self.inlet_conditions.V_app_infs)
        

    def get_Cxyz(self, wind : Wind, height_measure : float):
        # k number of sails, 1 (jb or main), 2 (jib and main)
        k = int(self.panels.shape[0] / (self.n_spanwise * self.n_chordwise * 2) ) # *2 in denominator due to underwater part
        sail_forces = np.split(self.force, (2*k))
        sail_areas = np.split(self.areas, (2*k))
        sails_Cxyz = []
        for i in range(k):
            Cxyz = coefs.get_vlm_Cxyz(sail_forces[i], np.array(wind.profile.get_true_wind_speed_at_h(height_measure)), self.rho, np.sum(sail_areas[i]))
            sails_Cxyz.append(Cxyz)
        return sails_Cxyz
            
        
    


import numpy as np
import sys
import os
import argparse
import shutil
import timeit

from pySailingVLM.rotations.csys_transformations import CSYS_transformations
from pySailingVLM.yacht_geometry.hull_geometry import HullGeometry
from pySailingVLM.results.save_utils import save_results_to_file
from pySailingVLM.solver.panels_plotter import display_panels_xyz_and_winds
from pySailingVLM.results.inviscid_flow import InviscidFlowResults
from pySailingVLM.solver.vlm import Vlm
from pySailingVLM.runner.sail import Wind, Sail
from pySailingVLM.solver.panels_plotter import plot_cp
from pySailingVLM.runner.container import Output, Rig, Conditions, Solver, MainSail, JibSail, Csys, Keel

def load_variable_module(args):
    try:
        sys.path.append(args.dvars)
        globals()['vr'] = __import__('variables')
    except ImportError as e:
        print(f'No variables.py file found in {args.dvars}\n{e}')
        sys.exit(0)
    
def parse_cli():

    parser = argparse.ArgumentParser(prog='pySailingVLM', description='Vortex lattice method for sailing')
    
    ##### required args
    requiredNamed = parser.add_argument_group('required arguments')
    #requiredNamed.add_argument('-v', '--vars', type=str, help="Directory containing input python file with variables, optional", default=os.getcwd(), required=False)   
    
    # dac z tego full path in relative

    # optional args
    # dir for storing output files
    #parser.add_argument('-d', '--dir', nargs='?', default=os.getcwd(), type=str, help="Alternative directory for for results, optional", required=False)
    parser.add_argument('-dv', '--dvars', type=str, help="Directory containing input python file with variables, optional", default=os.getcwd(), required=False)   
    # check if dv contain module!
    args = parser.parse_args()
    load_variable_module(args)
    return args                      


def main():
    #start = timeit.default_timer()
    parse_cli()
    out = Output(**vr.output_args)
    conditions = Conditions(**vr.conditions_args)
    solver = Solver(**vr.solver_args)
    main = MainSail(**vr.main_sail_args)
    jib = JibSail(**vr.jib_sail_args)
    csys = Csys(**vr.csys_args)
    keel = Keel(**vr.keel_args)
    rig = Rig(**vr.rig_args)
    
    csys_transformations = CSYS_transformations(
        conditions.heel_deg, conditions.leeway_deg,
        v_from_original_xyz_2_reference_csys_xyz=csys.reference_level_for_moments)

    
    w = Wind(conditions)
    s = Sail(solver, rig, main, jib, csys_transformations)
    sail_set = s.sail_set
    hull = HullGeometry(rig.sheer_above_waterline, rig.foretriangle_base, csys_transformations, keel.center_of_lateral_resistance_upright)
    myvlm = Vlm(sail_set.panels, solver.n_chordwise, solver.n_spanwise, conditions.rho, w.profile, sail_set.trailing_edge_info, sail_set.leading_edge_info)

    inviscid_flow_results = InviscidFlowResults(sail_set, csys_transformations, myvlm)
    inviscid_flow_results.estimate_heeling_moment_from_keel(hull.center_of_lateral_resistance)

    
    print(f"-------------------------------------------------------------")
    print(f"Notice:\n"
          f"\tThe forces [N] and moments [Nm] are without profile drag.\n"
          f"\tThe the _COG_ CSYS is aligned in the direction of the yacht movement (course over ground).\n"
          f"\tThe the _COW_ CSYS is aligned along the centerline of the yacht (course over water).\n"
          f"\tNumber of panels (sail sail_set with mirror): {sail_set.panels.shape}")

    df_components, df_integrals, df_inlet_IC = save_results_to_file(myvlm, csys_transformations, inviscid_flow_results, sail_set, out.name, out.file_name)
    shutil.copy(os.path.join(out.case_dir, out.case_name), os.path.join(out.name, out.case_name))

    print(df_integrals)
    #print(f"\nCPU time: {float(timeit.default_timer() - start):.2f} [s]")
    
    print("Preparing visualization.")   
    display_panels_xyz_and_winds(myvlm, inviscid_flow_results, myvlm.inlet_conditions, hull, show_plot=True, show_induced_wind=False, is_sailopt_mode=False)
    
    
    sails_Cxyz = myvlm.get_Cxyz(w, 1.0)

    print(f"Cxyz for {rig.sails_def}")
    for idx, Cxyz in enumerate(sails_Cxyz):
        print(f"C[{idx}]: {Cxyz}")
        a_vlm = Cxyz[1] / np.deg2rad(conditions.alpha_true_wind_deg)
        print(f"a_vlm[{idx}]: {a_vlm}")

    plot_cp(sail_set.zero_mesh, myvlm.p_coeffs, out.name)
    


import numba
import numpy as np
#from numpy.linalg import norm
from typing import Tuple


@numba.jit(nopython=True, cache=True, debug=True)
def is_in_vortex_core(vector_list : numba.typed.List) -> bool:
    """
    is_in_vortex_core check if list of vectors is n vortex core

    :param numba.typed.List vector_list: list of vectors
    :return bool: True or False
    """

    for vec in vector_list:
        if np.linalg.norm(vec) < 1e-8:
            return True
    return False


@numba.jit(numba.float64[::1](numba.float64[::1], numba.float64[::1], numba.float64[::1], numba.optional(numba.float64)), nopython=True, debug = True, cache=True) 
def vortex_line(p: np.array, p1: np.array, p2: np.array, gamma: float = 1.0) -> np.array:
    """
    vortex_line vortex line by Katz & Plotkin p 254

    :param np.array p: point at which  calculation is done
    :param np.array p1: point 1
    :param np.array p2: point 2
    :param float gamma: gamma orientation, defaults to 1.0
    :return np.array: velocity component
    """
    # strona 254

    r0 = np.asarray(p2 - p1)
    r1 = np.asarray(p - p1)
    r2 = np.asarray(p - p2)
    
    r1_cross_r2 = np.cross(r1, r2)
    
    q_ind = np.array([0.0, 0.0, 0.0], dtype=np.float64)
    # in nonpython mode must be list reflection to convert list to non python type
    # nested python oject can be badly converted -> recommend to use numba.typed.List
    # if r1 or r2 or |r1_cross_r2|^2 < epsilon
    # convert float - |r1_cross_r2|^2 to array with 1 element
    # this is due to numba
    # numba do not understand typed list with 2 vectors (r1 nad r2) and scalar like float
    sq = np.array([np.square(np.linalg.norm(r1_cross_r2))])
    b = is_in_vortex_core(numba.typed.List([r1, r2, sq]))

    if b:
        return np.asarray([0.0, 0.0, 0.0], dtype=np.float64)
    else:
        q_ind = r1_cross_r2 / np.square(np.linalg.norm(r1_cross_r2))
        q_ind *= np.dot(r0, (r1 / np.linalg.norm(r1) - r2 / np.linalg.norm(r2)))
        q_ind *= gamma / (4 * np.pi)

    return q_ind


@numba.jit(numba.float64[::1](numba.float64[::1], numba.float64[::1], numba.float64[::1], numba.optional(numba.float64)), nopython=True, debug = True, cache=True) 
# ten sposob jest wolniejszy niz wskazanie wprost typow
#@numba.jit(nopython=True, cache=True)
def vortex_infinite_line(P: np.ndarray, A: np.array, r0: np.ndarray, gamma : float = 1.0) -> np.ndarray:
    """
    vortex_infinite_line vortex infinite line

    :param np.ndarray P: point P
    :param np.array A: point A
    :param np.ndarray r0: r0 vector
    :param float gamma: gamma orientation, defaults to 1.0
    :return np.ndarray: velocity component
    """

    u_inf = r0 / np.linalg.norm(r0)
    ap = P - A
    norm_ap = np.linalg.norm(ap)

    v_ind = np.cross(u_inf, ap) / (
                norm_ap * (norm_ap - np.dot(u_inf, ap)))  # todo: consider checking is_in_vortex_core
    v_ind *= gamma / (4. * np.pi)
    return v_ind


@numba.jit(numba.float64[::1](numba.float64[::1], numba.float64[::1], numba.float64[::1], numba.float64[::1], numba.optional(numba.float64)), nopython=True, debug = True, cache=True) 
def vortex_horseshoe(p: np.array, B: np.array, C: np.array, V_app_infw: np.ndarray,
                        gamma: float = 1.0) -> np.array:
    """
    vortex_horseshoe _summary_

    :param np.array p: point form which calculation is done
    :param np.array B: point B
    :param np.array C: point C
    :param np.ndarray V_app_infw: apparent wind velocity for infinite sail
    :param float gamma: gamma orientation, defaults to 1.0
    :return np.array: velocity component
    
    
    B ------------------ +oo
    |
    |
    C ------------------ +oo
    """
    
    sub1 = vortex_infinite_line(p, C, V_app_infw, gamma)
    sub2 = vortex_line(p, B, C, gamma)
    sub3 = vortex_infinite_line(p, B, V_app_infw, -1.0 * gamma)
    q_ind = sub1 + sub2 + sub3
    return q_ind

# @numba.jit(nopython=True, cache=True) daje ten sam wynik co:
@numba.jit(numba.float64[::1](numba.float64[::1], numba.float64[::1], numba.float64[::1], numba.float64[::1], numba.float64[::1], numba.float64[::1], numba.optional(numba.float64)), nopython=True, debug = True, cache=True) 
def vortex_horseshoe_v2(p: np.array, A: np.array, B: np.array, C: np.array, D: np.array, V_app_infw : np.ndarray,
                gamma: float = 1.0) -> np.array:
    """
    vortex_ring vortex ring

    :param np.array p: point form which calculation is done
    :param np.array A: point A
    :param np.array B: point B
    :param np.array C: point C
    :param np.array D: point D
    :param np.ndarray V_app_infw: apparent wind velocity for infinite sail
    :param float gamma: gamma orientation, defaults to 1.0
    :return np.array: velocity component
    """
    sub1 = vortex_line(p, A, B, gamma)
    sub2 = vortex_line(p, B, C, gamma)
    sub3 = vortex_line(p, C, D, gamma)
    #sub4 = vortex_line(p, D, A, gamma)
    # wczesniej od C teraz od D
    inf_line_1 = vortex_infinite_line(p, D, V_app_infw, gamma)
    # wczesniej od B teraz od A
    inf_line_2 = vortex_infinite_line(p, A, V_app_infw, -1.0 * gamma)
    
    q_ind = sub1 + sub2 + sub3 + inf_line_1 + inf_line_2
    return q_ind


# @numba.jit(nopython=True, cache=True) daje ten sam wynik co:
@numba.jit(numba.float64[::1](numba.float64[::1], numba.float64[::1], numba.float64[::1], numba.float64[::1], numba.float64[::1], numba.optional(numba.float64)), nopython=True, debug = True, cache=True) 
def vortex_ring(p: np.array, A: np.array, B: np.array, C: np.array, D: np.array,
                gamma: float = 1.0) -> np.array:
    """
    vortex_ring vortex ring

    :param np.array p: point form which calculation is done
    :param np.array A: point A
    :param np.array B: point B
    :param np.array C: point C
    :param np.array D: point D
    :param float gamma: gamma orientation, defaults to 1.0
    :return np.array: velocity component
    """
    sub1 = vortex_line(p, A, B, gamma)
    sub2 = vortex_line(p, B, C, gamma)
    sub3 = vortex_line(p, C, D, gamma)
    sub4 = vortex_line(p, D, A, gamma)

    q_ind = sub1 + sub2 + sub3 + sub4
    return q_ind


# parallel - no time change
numba.jit(numba.float64[:, ::1](numba.float64[:, :, ::1], numba.float64[::1]), nopython=True, debug = True, cache=True) 
#@numba.jit(numba.float64[:, ::1](numba.float64[:, ::1], numba.float64[::1]), nopython=True, debug = True, cache=True) 
#@numba.jit(nopython=True, cache=True)
def calc_induced_velocity(v_ind_coeff : np.ndarray, gamma_magnitude : float) -> np.ndarray:
    """
    calc_induced_velocity calculate induced velocity

    :param np.ndarray v_ind_coeff: velocity induced ceofs
    :param float gamma_magnitude: gamma magnitude
    :return np.ndarray: induced wind velocity
    """
    N = gamma_magnitude.shape[0]

    V_induced = np.zeros((N, 3))
    for i in range(N):
        for j in range(N):
            V_induced[i] += v_ind_coeff[i,j] * gamma_magnitude[j]

    return V_induced


def calculate_app_fs(V_app_infs : np.ndarray, v_ind_coeff : np.ndarray, gamma_magnitude : float) -> Tuple[np.ndarray, np.ndarray]:
    """
    calculate_app_fs calculate apparent wind velocity for finite sail

    :param np.ndarray V_app_infs: apparent wind velocity for infinite sail
    :param np.ndarray v_ind_coeff: velocity induced coefs
    :param float gamma_magnitude: gamma magnitude
    :return Tuple[np.ndarray, np.ndarray]: wind velocity induced, wind velocity apparent finite sail
    """
    V_induced = calc_induced_velocity(v_ind_coeff, gamma_magnitude)
    V_app_fs = V_app_infs + V_induced
    return V_induced, V_app_fs






import pandas as pd
import os
import numpy as np
from pySailingVLM.results.inviscid_flow import InviscidFlowResults

from typing import List
from pySailingVLM.inlet.inlet_conditions import InletConditions
from pySailingVLM.yacht_geometry.sail_geometry import SailSet

from pySailingVLM.solver.additional_functions import get_cp_z_as_girths_all, get_cp_z_as_girths_all_above

def save_results_to_file(myvlm, csys_transformations,
                         inviscid_flow_results: InviscidFlowResults,
                         sail_set: SailSet,
                         output_dir="output", file_name='results'):
    cp_z_as_girths_all, cp_straight_yacht_all = get_cp_z_as_girths_all(sail_set, csys_transformations, myvlm.cp)
    cp_z_as_girths_all_above, names_all_above = get_cp_z_as_girths_all_above(cp_z_as_girths_all, sail_set)

    # take every second element
    n = 2 * len(sail_set.sails)
    a = tuple(np.array_split(cp_straight_yacht_all, n)[::2])
    cp_straight_yacht_above = np.concatenate(a) 
    
    df_inlet_conditions = np.array_split(myvlm.inlet_conditions.to_df_full(sail_set.sails[0].csys_transformations), 2)
    df_inlet_conditions_above = df_inlet_conditions[0]
    
    df_inviscid_flow = np.array_split(inviscid_flow_results.to_df_full(), 2)
    df_inviscid_flow_above = df_inviscid_flow[0]

    df_cp_straight_yacht_above = pd.DataFrame(cp_straight_yacht_above[:, 2], columns=['cp_points_upright.z'])
    df_cp_z_as_girths_all_above = pd.DataFrame(cp_z_as_girths_all_above, columns=['girths'])
    df_names_all_above = pd.DataFrame(names_all_above, columns=['sail_name'])

    list_of_df = [df_inviscid_flow_above, df_inlet_conditions_above,df_cp_straight_yacht_above, df_cp_z_as_girths_all_above, df_names_all_above ]

    df_inviscid_flow_integral = inviscid_flow_results.to_df_integral()
    df_inlet_conditions_integral = myvlm.inlet_conditions.winds.to_df_integral(sail_set.sails[0].csys_transformations)
    
    file_name = file_name + '.xlsx'
    df_merged = save_to_excel(output_dir, list_of_df, file_name, df_inviscid_flow_integral, df_inlet_conditions_integral)
   
    return df_merged, df_inviscid_flow_integral, df_inlet_conditions_integral
    


def save_to_excel(output_dir : str, list_of_df : List[pd.DataFrame], output_file_name : str, df_inviscid_flow_integral : pd.DataFrame, df_inlet_conditions_integral: pd.DataFrame):
    
    df_merged = pd.concat(list_of_df, axis=1, sort=False)
    max_row = df_merged.shape[0]
    last_col = df_merged.shape[1]
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    with pd.ExcelWriter(os.path.join(output_dir, output_file_name)) as writer:
        df_merged.to_excel(writer, startrow=0, sheet_name='Components', float_format="%.8f", index=False)

        workbook = writer.book
        worksheet = writer.sheets['Components']

        def merge_two_dicts(x, y):
            z = x.copy()  # start with keys and values of x
            z.update(y)  # modifies z with keys and values of y
            return z

        #Add some cell formats. https://xlsxwriter.readthedocs.io/example_pandas_column_formats.html
        general_format_dict = {'num_format': '#0.0000',
                               'border': 1,
                               'border_color': '#E8E8E8'  # grey
                               }

        format_general = workbook.add_format(general_format_dict)
        # https://xlsxwriter.readthedocs.io/format.html
        # https://htmlcolorcodes.com/color-picker/
        format_COG = workbook.add_format(
            merge_two_dicts(general_format_dict,
                            {'bg_color': '#FDF8E1'  # light sand
                            }))

        format_COW = workbook.add_format(
            merge_two_dicts(general_format_dict,
                            {'bg_color': '#EAF4FF'  # light sand
                            }))

        format_percent = workbook.add_format(
            {'num_format': '0.00%',
             'border': 1,
             'border_color': '#E8E8E8'  # grey
            })

        for i, col in enumerate(df_merged.columns):
            column_len = df_merged[col].astype(str).str.len().max()
            # Setting the length if the column header is larger than the max column value length
            column_len = max(column_len, len(col)) + 2  # + padding

            worksheet.set_column(i, i, column_len, cell_format=format_general)
            
            if "COG" in col:
                # workaround because set_column finction do not allow to do range cell formatting
                # first row, first col, last row, last col
                worksheet.conditional_format(1, i, max_row, i, {'type': 'cell','criteria': '>=','value': 0, 'format': format_COG})
                worksheet.conditional_format(1, i, max_row, i, {'type': 'cell','criteria': '<','value': 0, 'format': format_COG})

            if "COW" in col:
                worksheet.conditional_format(1, i, max_row, i, {'type': 'cell','criteria': '>=','value': 0, 'format': format_COW})
                worksheet.conditional_format(1, i, max_row, i, {'type': 'cell','criteria': '<','value': 0, 'format': format_COW})

            if col == 'Camber_estimate':
                worksheet.conditional_format(1, i, max_row, i, {'type': 'cell','criteria': '>=','value': 0, 'format': format_percent})
                worksheet.conditional_format(1, i, max_row, i, {'type': 'cell','criteria': '<','value': 0, 'format': format_percent})

        def write_summary_sheet(df, sheet_name):
            df.to_excel(writer, startrow=0, sheet_name=sheet_name, float_format="%.2f", index=False)
            worksheet = writer.sheets[sheet_name]
            for i, col in enumerate(df.columns):
                column_len = df[col].astype(str).str.len().max()
                # Setting the length if the column header is larger than the max column value length
                column_len = max(column_len, len(str(col))) + 0  # + padding
                worksheet.set_column(i, i, column_len, cell_format=format_general)

            for i, row in df.iterrows():
                ri = i + 1
                if "COG" in row['Quantity']:
                    # workaround format first column with string if are not blank (they will always containg string)
                    worksheet.conditional_format(1, 0, ri, 0, {'type':  'no_blanks', 'format': format_COG})
                    worksheet.conditional_format(0, 1, ri, 1, {'type': 'cell','criteria': '>=','value': 0, 'format': format_COG})
                    worksheet.conditional_format(0, 1, ri, 1, {'type': 'cell','criteria': '<','value': 0, 'format': format_COG})

                if "COW" in row['Quantity']:
                    worksheet.conditional_format(1, 0, ri, 0, {'type':  'no_blanks', 'format': format_COW})
                    worksheet.conditional_format(0, 1, ri, 1, {'type': 'cell','criteria': '>=','value': 0, 'format': format_COW})
                    worksheet.conditional_format(0, 1, ri, 1, {'type': 'cell','criteria': '<','value': 0, 'format': format_COW})


        write_summary_sheet(df_inviscid_flow_integral, 'Integrals')
        write_summary_sheet(df_inlet_conditions_integral, 'IC_at_reference_height')

        return df_merged


from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pySailingVLM.rotations.csys_transformations import CSYS_transformations
    from pySailingVLM.yacht_geometry.sail_geometry import SailSet

import numpy as np
import numba
import matplotlib.pyplot as plt
from typing import List, Tuple

from pySailingVLM.solver.velocity import vortex_ring, vortex_horseshoe




#@numba.jit(numba.float64[::1](numba.float64[::1]), nopython=True, debug=False)
def normalize(x):
    # xn = x / norm(x)
    xn = x / np.linalg.norm(x)
    return xn

def extract_above_water_quantities(quantities : np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """
    extract_above_water_quantities get data above water

    :param np.ndarray quantities: data to be "searched" for above water quantities
    :return Tuple[np.ndarray, np.ndarray]: above water quantities, total above water quantities
    """
    
    # for jib and main, quantities is always dividable by 2
    half = int(quantities.shape[0] / 2)
    above_water_quantities = quantities[0:half]
    total_above_water_quantities = np.sum(above_water_quantities, axis=0)  # heeling, sway, yaw (z-axis)
    return above_water_quantities, total_above_water_quantities


# circular import
#  csys_transformations: CSYS_transformations
def get_cp_strainght_yacht(cp_points : np.ndarray, csys_transformations : CSYS_transformations) -> np.ndarray:
    """
    cp_strainght_yacht get center of pressure points straight to bridge

    :param np.ndarray cp_points: center of pressure points
    :param CSYS_transformations: : CSYS_transformations objetc
    :return np.ndarray: staright center of pressure points
    """
    return np.array([csys_transformations.reverse_rotations_with_mirror(p) for p in cp_points])

def cp_to_girths(sail_cp_straight_yacht : np.ndarray, tack_mounting : np.ndarray) -> np.ndarray:
    """
    cp_to_girths convert center of pressure points to girths

    :param np.ndarray sail_cp_straight_yacht: straighten yacht center of pressure points
    :param np.ndarray tack_mounting: tack mounting
    :return np.ndarray: center of pressure points as girths
    """
    
    sail_cp_straight_yacht_z = sail_cp_straight_yacht[:, 2]
    z_as_girths = (sail_cp_straight_yacht_z - tack_mounting[2]) / (max(sail_cp_straight_yacht_z) - tack_mounting[2])
    return z_as_girths

# circular import
#  csys_transformations: CSYS_transformations
# sail_set : SailSet problem with import
def get_cp_z_as_girths_all(sail_set : SailSet, csys_transformations : CSYS_transformations, center_of_pressure : np.ndarray) -> np.ndarray: 
    """
    cp_z_as_girths_all get center of pressure staright z as girths

    :param SailSet sail_set: Sail set object
    :param CSYS_transformations csys_transformations: csys transformations
    :param np.ndarray center_of_pressure: array with all center of pressure points (for all sails and for above and under water)

    :return np.ndarray: y_as_girths for all sails (with above and under water points)
    """
    n = len(sail_set.sails)
    # 2* bo mamy odpicie lustrzane 
    chunks_cp_points = np.array_split(center_of_pressure, 2 * n)

    nxm = chunks_cp_points[0].shape[0]
    cp_z_as_girths_all = np.array([])
    #cp_straight_yacht_all = np.array([])
    
    cp_straight_yacht_all = np.empty((0,3))
    for i in range(n):
        sail_cp_points = np.concatenate([chunks_cp_points[i], chunks_cp_points[i+n]])
        sail_cp_straight_yacht = get_cp_strainght_yacht(sail_cp_points, csys_transformations)
        z_as_girths = cp_to_girths(sail_cp_straight_yacht, sail_set.sails[i].tack_mounting)
        cp_z_as_girths_all = np.append(cp_z_as_girths_all, z_as_girths)
        #cp_straight_yacht_all = np.append(cp_straight_yacht_all, sail_cp_straight_yacht)
        cp_straight_yacht_all = np.append(cp_straight_yacht_all, sail_cp_straight_yacht, axis=0)

    # ulozenie w array cp_z_as_girths_all :
    # jib above, jib under, main abobe, main under
    return cp_z_as_girths_all, cp_straight_yacht_all


# to policzyc dla wszystkich cpsow
# potem wziac gorna czesc czyli pierwsza polowe by miec nad woda
# import problem
# sail set  sail_set : SailSet
def get_cp_z_as_girths_all_above(cp_z_as_girths_all : np.ndarray, sail_set : SailSet) -> Tuple[np.ndarray, np.ndarray]:
    """
    get_cp_z_as_girths_all_above get center of pressure z coordinate as girths above water

    :param np.ndarray cp_z_as_girths_all: center of pressure z coordinate as girths
    :param SailSet sail_set: sail set 
    :return Tuple[np.ndarray, np.ndarray]:  center of pressure z coordinate as girths above water for all panels, names of all panels
    """
    # 2 bo mamy odbicie 
    # parzyste numery to sa te nad woda
    n = 2 * len(sail_set.sails)
    z_splitted = np.split(cp_z_as_girths_all, n)
    # [::2] gets every second element from array z_splitted
    cp_z_as_girths_all_above = np.asarray(z_splitted[::2]).flatten()
    # cp_z_as_girths_all_above size
    repeat = int(cp_z_as_girths_all_above.shape[0] / len(sail_set.sails))
    names = np.array([])
    
    for sail in sail_set.sails:
        rep_names = np.repeat(sail.name, repeat)
        names = np.append(names, rep_names)
      
    return cp_z_as_girths_all_above, names      


def plot_mesh(mesh1 : np.array, mesh2 : np.array = None, show : bool = False, dimentions : List  = [0, 1, 2], color1 : str = 'green', color2 : str = None, title : str = '3d plot'):

    f = plt.figure(figsize=(12, 12))
    labels = {0: 'X', 1: 'Y', 2: 'Z'}
    ax = plt.axes(projection='3d')
    if len(dimentions) == 2:
        ax = plt.axes()
    
    ax.set_title(title)
    
    if len(dimentions) == 2:
        ax.set_xlabel(labels[dimentions[0]])
        ax.set_ylabel(labels[dimentions[1]])
    elif len(dimentions) == 3:
        ax.set_xlabel(labels[dimentions[0]])
        ax.set_ylabel(labels[dimentions[1]])
        ax.set_zlabel(labels[dimentions[2]])


    for i in range(mesh1.shape[0]):
        if len(dimentions) == 2:
            ax.plot(mesh1[i, :, dimentions[0]], mesh1[i, :, dimentions[1]], color1)
            if mesh2 is not None:
                ax.plot(mesh2[i, :, dimentions[0]], mesh2[i, :, dimentions[1]], color2)
        else:
            ax.plot3D(mesh1[i, :, 0], mesh1[i, :, 1], mesh1[i, :, 2], color1)
            if mesh2 is not None:
                ax.plot3D(mesh2[i, :, 0], mesh2[i, :, 1], mesh2[i, :, 2], color2)

    if show:
        plt.show()


import numpy as np
import os
import pandas as pd
from typing import List, Tuple
from numpy.linalg import norm

import numba

from pySailingVLM.solver.velocity import vortex_ring, vortex_horseshoe, vortex_horseshoe_v2

# parallel slows down beacuse loop is not big
@numba.jit(numba.types.Tuple((numba.float64[:, ::1], numba.float64[:, :, ::1])) (numba.float64[:, ::1], numba.float64[:, ::1], numba.float64[:, :, ::1], numba.float64[:, ::1], numba.boolean[::1], numba.float64), nopython=True, debug = True, cache=True)
# @numba.jit(numba.types.Tuple((numba.float64[:, ::1], numba.float64[:, ::1])) (numba.float64[:, ::1], numba.float64[:, ::1], numba.float64[:, :, ::1], numba.float64[:, ::1], numba.boolean[::1], numba.float64), nopython=True, debug = True, cache=True)
def calc_velocity_coefs(V_app_infw, points_for_calculations, rings, normals, trailing_edge_info : np.ndarray, gamma_orientation : np.ndarray):
    m = points_for_calculations.shape[0]

    coefs = np.zeros((m,m)) # coefs calculated for normalized velocity
    v_ind_coeff = np.zeros((m, m, 3))
    for i in range(points_for_calculations.shape[0]):

        # loop over other vortices
        for j in range(rings.shape[0]):
            A = rings[j][0]
            B = rings[j][1]
            C = rings[j][2]
            D = rings[j][3]
            
            a = vortex_ring(points_for_calculations[i], A, B, C, D, gamma_orientation)
            # poprawka na trailing edge
            # todo: zrobic to w drugim, oddzielnym ifie
            if trailing_edge_info[j]:
                #a = vortex_horseshoe(points_for_calculations[i], B, C, V_app_infw[j], gamma_orientation)
                a = vortex_horseshoe_v2(points_for_calculations[i], A, B, C, D,V_app_infw[j], gamma_orientation)
            coefs[i, j] = np.dot(a, normals[i].reshape(3, 1))[0]
            # this is faster than wind_coefs[i, j, :] = a around 0.1s (case 10x10)
            
            v_ind_coeff[i, j, 0] = a[0] # FOR NUMBA
            v_ind_coeff[i, j, 1] = a[1]
            v_ind_coeff[i, j, 2] = a[2]
            
    return coefs, v_ind_coeff

# do wywalenia pozniej
def get_leading_edge_mid_point(p2: np.ndarray, p3: np.ndarray) -> np.ndarray:
    return (p2 + p3) / 2.


# do wywalenia pozniej
def get_trailing_edge_mid_points(p1: np.ndarray, p4: np.ndarray) -> np.ndarray:
    return (p4 + p1) / 2.

def calculate_normals_collocations_cps_rings_spans_leading_trailing_mid_points(panels: np.ndarray, gamma_orientation : float) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    calculate_normals_collocations_cps_rings_spans_leading_trailing_mid_points _summary_

    :param np.ndarray panels: panels
    :param float gamma_orientation: gamma orientation
    :return Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]: normals, collocation points, center of pressure, rings, span vectors, leading mid points, trailing edge mid points
    """
    K = panels.shape[0]
    ns = np.zeros((K, 3))
    span_vectors = np.zeros((K, 3))
    ctr_p = np.zeros((K, 3))
    cp = np.zeros((K, 3))
    leading_mid_points = np.zeros((K, 3))
    trailing_edge_mid_points = np.zeros((K, 3))
    
    rings = np.zeros(shape=panels.shape)
    for idx, panel in enumerate(panels):
        p1 = panel[0]
        p2 = panel[1]
        p3 = panel[2]
        p4 = panel[3]

        # tutaj bylo zmienione!!!
        #vect_A = p4 - p2
        #vect_B = p3 - p1
        
        vect_B = p4 - p2
        vect_A = p3 - p1

        leading_mid_points[idx] = get_leading_edge_mid_point(p2, p3)
        trailing_edge_mid_points[idx] = get_trailing_edge_mid_points(p1, p4)
        dist = trailing_edge_mid_points[idx] - leading_mid_points[idx]

        ctr_p[idx] = leading_mid_points[idx] + 0.75 * dist
        cp[idx] = leading_mid_points[idx] + 0.25 * dist

        p2_p1 = p1 - p2
        p3_p4 = p4 - p3

        A = p1 + p2_p1 / 4.
        B = p2 + p2_p1 / 4.
        C = p3 + p3_p4 / 4.
        D = p4 + p3_p4 / 4.

        rings[idx] = np.array([A, B, C, D])
        
        # span vectors
        bc = C - B
        bc *= gamma_orientation
        span_vectors[idx] = bc
        n = np.cross(vect_A, vect_B)
        n = n / np.linalg.norm(n)
        ns[idx] = n
    return ns, ctr_p, cp, rings, span_vectors, leading_mid_points, trailing_edge_mid_points

def calculate_vlm_variables(panels: np.ndarray, trailing_edge_info : np.ndarray, gamma_orientation : float, n_chordwise : int, n_spanwise: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    calculaye_vlm_variables calculate normals, control points, center of pressure, rings, span vectors, leading mid points, trailing edge mid points

    :param np.ndarray panels: panels
    :param float gamma_orientation: gamma orientation
    :param int n_chordwise: number of panels chordwise
    :param int n_spanwise: number of panels spanwise
    :return Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]: normals, collocation points, center of pressure, rings, span vectors, leading mid points, trailing edge mid points
    """
    K = panels.shape[0]
    G = int (K / (n_chordwise * n_spanwise))
    panels_splitted = np.array_split(panels, G)
    trailing_edge_info_splitted = np.array_split(trailing_edge_info, G)
    
    rings = np.zeros(shape=(G, n_chordwise * n_spanwise, 4, 3))
    ns = np.zeros(shape=(G, n_chordwise * n_spanwise, 3))
    span_vectors = np.zeros((G, n_chordwise * n_spanwise, 3))
    ctr_p = np.zeros((G, n_chordwise * n_spanwise, 3))
    cp = np.zeros((G, n_chordwise * n_spanwise, 3))
    leading_mid_points = np.zeros((G, n_chordwise * n_spanwise, 3))
    trailing_edge_mid_points = np.zeros((G, n_chordwise * n_spanwise, 3))
    
    for idx, (grid_panels, trailing_grid_info) in enumerate(zip(panels_splitted,trailing_edge_info_splitted)):

        n_panels = grid_panels.shape[0]
        for k in range(n_panels):

            currnet_panel = grid_panels[k]
            c_p1 = currnet_panel[0]
            c_p2 = currnet_panel[1]
            c_p3 = currnet_panel[2]
            c_p4 = currnet_panel[3]
            
            vect_B = c_p4 - c_p2
            vect_A = c_p3 - c_p1

            leading_mid_points[idx, k] = (c_p2 + c_p3) / 2.0 
            trailing_edge_mid_points[idx, k] = (c_p1 + c_p4) / 2.0 
            dist = trailing_edge_mid_points[idx, k] - leading_mid_points[idx, k]

            ctr_p[idx, k] = leading_mid_points[idx, k] + 0.75 * dist
            cp[idx, k] = leading_mid_points[idx, k] + 0.25 * dist

            c_p2_p1 = c_p1 - c_p2
            c_p3_p4 = c_p4 - c_p3
            
            B = c_p2 + c_p2_p1 / 4.
            C = c_p3 + c_p3_p4 / 4.
                
            n = np.cross(vect_A, vect_B)
            n = n / np.linalg.norm(n)
            ns[idx, k] = n
            
            if trailing_grid_info[k]:  
                # if trailing edge
                A = c_p1 + c_p2_p1 / 4.
                D = c_p4 + c_p3_p4 / 4.
            else:
                # if not trailing edge
                next_panel = grid_panels[k + n_spanwise]
                n_p1 = next_panel[0]
                n_p2 = next_panel[1]
                n_p3 = next_panel[2]
                n_p4 = next_panel[3]

                n_p2_p1 = n_p1 - n_p2
                n_p3_p4 = n_p4 - n_p3
                
                n_B = n_p2 + n_p2_p1 / 4.
                n_C = n_p3 + n_p3_p4 / 4.
            
                A = n_B
                D = n_C
                
            rings[idx, k] = np.array([A, B, C, D])
            # span vectors
            bc = C - B
            bc *= gamma_orientation
            span_vectors[idx,k] = bc


    span_vectors = span_vectors.reshape(K,3)
    rings = rings.reshape(K, 4, 3)

    ns =ns.reshape(K,3)

    ctr_p = ctr_p.reshape(K,3)
    cp = cp.reshape(K,3)
    leading_mid_points = leading_mid_points.reshape(K,3)
    trailing_edge_mid_points = trailing_edge_mid_points.reshape(K,3)         
    
    return ns, ctr_p, cp, rings, span_vectors, leading_mid_points, trailing_edge_mid_points


def calculate_RHS(V_app_infw, normals):
    RHS = -V_app_infw.dot(normals.transpose()).diagonal()
    RHS = np.asarray(RHS)
    return RHS


def solve_eq(coefs: np.ndarray, RHS: np.ndarray) -> np.ndarray:
    """
    solve_eq solve equation for gamma 

    :param np.ndarray coefs: coefs
    :param np.ndarray RHS: RHS
    :return np.ndarray: calculated gamma
    """
    big_gamma = np.linalg.solve(coefs, RHS)
    return big_gamma


# do przykladu z aircraftem
def get_vlm_CL_CD_free_wing(F: np.ndarray, V: np.array, rho : float, S : float) -> Tuple[float, float]:
    
    total_F = np.sum(F, axis=0)
    q = 0.5 * rho * (np.linalg.norm(V) ** 2) * S
    CL_vlm = total_F[2] / q
    CD_vlm = total_F[0] / q
    
    return CL_vlm, CD_vlm

def get_vlm_Cxyz(F: np.ndarray, V: np.array, rho : float, S : float) -> Tuple[float, float, float]:
    
    total_F = np.sum(F, axis=0)
    q = 0.5 * rho * (np.linalg.norm(V) ** 2) * S
    #Cx_vlm, Cy_vlm, Cz_vlm = total_F / q
    
    return total_F / q #Cx_vlm, Cy_vlm, Cz_vlm

def get_CL_CD_free_wing(AR, AoA_deg):
    #  TODO allow tapered wings in book coeff_formulas
    a0 = 2. * np.pi  # dCL/d_alfa in 2D [1/rad]
    e_w = 0.8  # span efficiency factor, range: 0.8 - 1.0

    a = a0 / (1. + a0 / (np.pi * AR * e_w))

    CL_expected_3d = a * np.deg2rad(AoA_deg)
    CD_ind_expected_3d = CL_expected_3d * CL_expected_3d / (np.pi * AR * e_w)

    return CL_expected_3d, CD_ind_expected_3d




import numpy as np

from pySailingVLM.solver.coefs import calc_velocity_coefs
from pySailingVLM.solver.velocity import calculate_app_fs

def calc_moment_arm_in_shifted_csys(cp_points, v_from_old_2_new_csys):
    dx, dy, dz = v_from_old_2_new_csys
    N = len(cp_points)
    r = np.copy(cp_points)
    for i in range(N):
        r[i, 0] -= dx  # yawing moment arm
        r[i, 1] -= dy  # pitching moment arm
        if cp_points[i, 2] > 0:
            r[i, 2] -= dz  # heeling moment arm
        else:
            r[i, 2] += dz

    return r

def calc_moments(cp_points, forces):
    moments = np.cross(cp_points, forces)
    # Mx_test1 = cp_points[:, 2] * force_xyz[:, 1]
    # Mx_test2 = cp_points[:, 1] * force_xyz[:, 2]
    # Mx_test = Mx_test1 + Mx_test2
    return moments

def determine_vector_from_its_dot_and_cross_product(F, r_dot_F, r_cross_F):
    # https://math.stackexchange.com/questions/246594/what-is-vector-division
    # https://math.stackexchange.com/questions/1683996/determining-an-unknown-vector-from-its-cross-and-dot-product-with-known-vector
    # we know that: a x ( b x c) = b(a . c) - c(a.b)
    # so: F x ( r x F) = r(F . F) - F(F.r)
    # one shall get r, assuming that both the cross and dot product are known:
    # r = (F x (r x F) + F(F . r))/(F . F)

    F_dot_F = np.dot(F, F)
    R = (np.cross(F, r_cross_F) + F * r_dot_F) / F_dot_F
    return np.array(R)




# to bylo w starej finkcji:
# # "Loop serialization occurs when any number of prange driven loops are present 
# # inside another prange driven loop. In this case the outermost of all the prange loops executes
# # in parallel and any inner prange loops (nested or otherwise) 
# # are treated as standard range based loops. Essentially, nested parallelism does not occur."
# #@numba.jit(nopython=True, parallel=True)

# cp - center of pressure
def calc_force_wrapper(V_app_infw, gamma_magnitude, rho, cp, rings, n_spanwise, normals, span_vectors, trailing_edge_info : np.ndarray, leading_edges_info : np.ndarray, gamma_orientation : float = 1.0):
    # Katz and Plotkin, p. 346 Chapter 12 / Three-Dimensional Numerical Solution
    # f. Secondary Computations: Pressures, Loads, Velocities, Etc
    # Eq (12.25)
    _, v_ind_coeff = calc_velocity_coefs(V_app_infw, cp, rings, normals, trailing_edge_info, gamma_orientation)
    V_induced, V_at_cp = calculate_app_fs(V_app_infw, v_ind_coeff, gamma_magnitude)
    
    # if case 1x1 leading_edges_info is False False False False
    # horseshoe_edge_info i True True True True
    # caclulating winds as for trailing edges
    # forces and pressure like "leading edge"
    case1x1 = np.logical_not(np.any(leading_edges_info)) 
    
    K = cp.shape[0]
    force_xyz = np.zeros((K, 3))
    #numba.prange
    for i in range(K):
        # for spanwise only!
        # if panel is leading edge
        gamma = 0.0
        if leading_edges_info[i] or case1x1:
            gamma = span_vectors[i] * gamma_magnitude[i]
        else:
            gamma = span_vectors[i] * (gamma_magnitude[i] - gamma_magnitude[i-n_spanwise])
        force_xyz[i] = rho * np.cross(V_at_cp[i], gamma)
        
    return force_xyz, V_at_cp, V_induced



def calc_pressure(forces, normals, areas):
    p = forces.dot(normals.transpose()).diagonal() /  areas
    return p


def is_no_flux_BC_satisfied(V_app_fw, panels, areas, normals):

    flux_through_panel = -V_app_fw.dot(normals.transpose()).diagonal()

    for area in areas:
        if np.isnan(area) or area < 1E-14:
            raise ValueError("Solution error, panel_area is suspicious")

    for flux in flux_through_panel:
        if abs(flux) > 1E-12 or np.isnan(flux):
            raise ValueError("Solution error, there shall be no flow through panel!")

    return True

def calc_pressure_coeff(pressure: np.ndarray, rho: float, V: np.ndarray) -> np.ndarray:
    """
    calc_pressure_coeff calculate pressure coeffs

    :param np.ndarray pressure: array with pressure
    :param float rho: liquid density
    :param np.ndarray V: free stream speed
    :return np.ndarray: pressure coeffs
    """
    # 
    p_coeffs = np.array([ p / (0.5 * rho * np.dot(V[i], V[i])) for i, p in enumerate(pressure)])
    return p_coeffs
    


import numpy as np

from pySailingVLM.solver import mesher
from pySailingVLM.rotations import geometry_calc

from typing import List, Tuple
import numba


def make_panels_from_mesh_spanwise(mesh : np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    make_panels_from_mesh_spanwise make panels from spanwise mesh

    :param np.ndarray mesh: mesh array
    :return Tuple[np.ndarray, np.ndarray, np.ndarray]: panels, trailing_edge_info, leading_edge_info
    """
    n_lines = mesh.shape[0]
    n_points_per_line = mesh.shape[1]
    M = n_points_per_line - 1
    N = n_lines - 1
    
    panels = np.zeros((N * M, 4, 3))
    trailing_edge_info = np.full(N * M, False, dtype=bool)
    leading_edge_info = np.full(N * M, False, dtype=bool)
    
    counter = 0
    for i in range(N):
        for j in range(M):
            pSE = mesh[i + 1][j]
            pSW = mesh[i][j]
            pNW = mesh[i][j + 1]
            pNE = mesh[i + 1][j + 1]
            panels[counter] = [pSE, pSW, pNW, pNE] 

            if i == (n_lines - 2):
                trailing_edge_info[counter] = True
                
            else:
                if i == 0:
                    leading_edge_info[counter] = True
                
            counter += 1
            
    return  panels, trailing_edge_info, leading_edge_info

# for aircraft
def make_panels_from_le_te_points(points : np.ndarray, grid_size :List[int]) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    make_panels_from_le_te_points make_panels_from_le_te_points make panels from leading end trailing edges points


    :param np.ndarray points: points
    :param List[int, int] grid_size: grid size
    :return Tuple[np.ndarray, np.ndarray, np.ndarray]: panels, trailing edge info, leading edge info
    """

    le_SW, te_SE, le_NW, te_NE = points
    nc, ns = grid_size
    south_line = mesher.discrete_segment(le_SW, te_SE, nc)
    north_line = mesher.discrete_segment(le_NW, te_NE, nc)

    mesh = mesher.make_point_mesh(south_line, north_line, ns)
    panels, trailing_edge_info, leading_edge_info = make_panels_from_mesh_spanwise(mesh)
    return panels, trailing_edge_info, leading_edge_info

# not used??
# def make_panels_from_le_points_and_chords(le_points : List[np.ndarray], grid_size : List[int], chords: np.ndarray, interpolated_camber: np.ndarray, interpolated_distance_from_luff: np.ndarray, gamma_orientation : float) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
#     """
#     make_panels_from_le_points_and_chords make panels from leading edges and chords

#     :param List[np.ndarray] le_points: leading edges points
#     :param List[int] grid_size: zise of grid
#     :param np.ndarray chords: chords 
#     :param np.ndarray: interpolated_camber
#     :param np.ndarray: interpolated_distance_from_luff
#     :param float gamma_orientation: gamma orientation
#     :return Tuple[np.ndarray, np.ndarray, np.ndarray]: panels, trailing_edge_info, leading_edge_info
#     """
    
#     le_SW,  le_NW = le_points
#     n_chordwise, n_spanwise = grid_size

    
#     chords_vec = np.array([chords, np.zeros(len(chords)), np.zeros(len(chords))])
#     chords_vec = chords_vec.transpose()
#     fchords_vec = np.flip(chords_vec, axis=0)
#     mesh = mesher.make_airfoil_mesh([le_SW, le_NW], grid_size, chords_vec, interpolated_distance_from_luff, interpolated_camber)
#     rmesh = np.array([self.csys_transformations.rotate_point_with_mirror(point) for point in mesh])
        
#     te_line = np.copy(le_line)  # deep copy
#     te_line += chords_vec
  
#     mesh = mesher.make_point_mesh(le_line, te_line, n_chordwise)

#     mesh = np.swapaxes(mesh, 0, 1)
#     panels, trailing_edge_info, leading_edge_info = make_panels_from_mesh_spanwise(mesh)
    
#     return panels, trailing_edge_info, leading_edge_info


# no seppedup
#@numba.jit(nopython=True, parallel=True)
def get_panels_area(panels: np.ndarray)-> np.ndarray:
    """
    get_panels_area get panels area

    :param np.ndarray panels: panels
    :return np.ndarray: areas
    """

    m = panels.shape[0]
    areas = np.zeros(m)

    sh = panels.shape[0]
    # numba.prange
    # range works slightly quicker than numba.prange (without decorator of course)
    for i in range(sh):
        
        p = [panels[i, 0], panels[i, 1], panels[i, 2], panels[i, 3]]
        path = []
        for j in range(len(p) - 1):
            step = p[j + 1] - p[j]
            path.append(step)

        area = 0
        for k in range(len(path) - 1):
            s = np.cross(path[k], path[k + 1])
            s = np.linalg.norm(s)
            area += 0.5 * s
        areas[i] = area   

    return areas


from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from pySailingVLM.runner.container import Rig, JibSail, MainSail, Solver, Conditions

import sys
import numpy as np
from dataclasses import dataclass

from typing import Literal, Optional
from pySailingVLM.inlet.winds import ExpWindProfile, FlatWindProfile, LogWindProfile
from pySailingVLM.rotations.csys_transformations import CSYS_transformations
from pySailingVLM.yacht_geometry.sail_factory import SailFactory
from pySailingVLM.yacht_geometry.sail_geometry import SailSet
from pySailingVLM.solver.interpolator import Interpolator
     
def check_list_for_none(list_of_different_obj: list):
    # list like: a = [np.array([1.9]), 1.0, None] should return true
    # a = [np.array([1.9]), 1.0] should return false
    res = False
    for element in list_of_different_obj:
        if element is None:
            res = True
            break
    return res

def check_input_arr_shapes(check_arr : list, sails_def: str):
    for arr in check_arr:
        if arr.shape != check_arr[0].shape:
            raise ValueError(f'Not all {sails_def.upper()} arrays are equal size!')    
@dataclass
class Wind:
    ## czy dataclass sprawdza type of input ???
    conditions: Conditions
    # alpha_true_wind_deg: float
    # tws_ref: float
    # SOG_yacht: float
    
    # wind_exp_coeff: float  = None # expotentional
    # wind_reference_measurment_height: float = None# expotentional or logarithmic
    # reference_water_level_for_wind_profile: float = None# expotentional
    # roughness: float = None # logarithmic
    # wind_profile: Literal['exponential', 'flat', 'logarithmic = 'exponential'
    
    def __post_init__(self):
        try: 
            if self.conditions.wind_profile == 'logarithmic' and None in [self.conditions.roughness, self.conditions.wind_reference_measurment_height]:
                raise ValueError("ERROR in logarithmic mode!: Roughness or wind_reference_measurment_height badly defined!")
            elif self.conditions.wind_profile == 'expotentional' and None in [self.conditions.wind_exp_coeff, self.conditions.wind_reference_measurment_height]:
                raise ValueError("ERROR in expotentional mode!: wind_exp_coeff or wind_reference_measurment_height badly defined!")
            elif self.conditions.wind_profile == 'flat' and None in [self.conditions.wind_exp_coeff, self.conditions.wind_reference_measurment_height]:
                raise ValueError("ERROR in flat mode!: wind_exp_coeff or wind_reference_measurment_height badly defined!")
        except ValueError as err:
            print(err)
            sys.exit()
            
    def __get_profile(self):
        if self.conditions.wind_profile == 'exponential':
            wind = ExpWindProfile(
                self.conditions.alpha_true_wind_deg, self.conditions.tws_ref, self.conditions.SOG_yacht,
                exp_coeff=self.conditions.wind_exp_coeff,
                reference_measurment_height=self.conditions.wind_reference_measurment_height,
                reference_water_level_for_wind_profile=self.conditions.reference_water_level_for_wind_profile)
        elif self.conditions.wind_profile == 'flat':
            wind = FlatWindProfile(self.conditions.alpha_true_wind_deg, self.conditions.tws_ref, self.conditions.SOG_yacht)
        else:
            wind = LogWindProfile(
                self.conditions.alpha_true_wind_deg, self.conditions.tws_ref, self.conditions.SOG_yacht,
                roughness=self.conditions.roughness,
                reference_measurment_height=self.conditions.wind_reference_measurment_height)
        return wind

    @property
    def profile(self):
        return self.__get_profile()
    
@dataclass
class Sail:
    
    solver: Solver
    rig: Rig
    main_sail: MainSail
    jib_sail: JibSail
    csys_transformations: CSYS_transformations
     
    def __post_init__(self):
        try:
            if self.jib_sail is not None:
                jib_stuff = [self.rig.foretriangle_base, self.rig.foretriangle_height, self.rig.jib_luff, self.jib_sail.girths, self.jib_sail.centerline_twist_deg, self.jib_sail.camber, self.jib_sail.camber_distance_from_luff]
            if self.main_sail is not None:
                main_stuff = [self.rig.main_sail_luff, self.main_sail.girths, self.main_sail.centerline_twist_deg, self.main_sail.camber, self.main_sail.camber_distance_from_luff]
            
            if self.rig.sails_def == 'jib' and check_list_for_none(jib_stuff): # if JIB and all vars are not NONE
                raise ValueError('Not all JIB variables are speciified!')
            elif self.rig.sails_def == 'main' and check_list_for_none(main_stuff):
                raise ValueError('Not all MAIN variables are speciified!') 
            elif self.rig.sails_def == 'jib_and_main' and check_list_for_none([*jib_stuff, *main_stuff]):
                raise ValueError('Not all JIB or MAIN variables are speciified!')
            
            if self.rig.sails_def == 'jib':
                check_input_arr_shapes(jib_stuff[3:], self.rig.sails_def)
            elif self.rig.sails_def =='main':
                check_input_arr_shapes(main_stuff[1:], self.rig.sails_def)
            else:
                check_input_arr_shapes(jib_stuff[3:], self.rig.sails_def)
                check_input_arr_shapes(main_stuff[1:], self.rig.sails_def)
        except ValueError as e:
            print(e)
            sys.exit()

    def __get_sail_set(self):
        
        interpolator = Interpolator(self.solver.interpolation_type)
        factory = SailFactory(csys_transformations=self.csys_transformations, n_spanwise=self.solver.n_spanwise, n_chordwise=self.solver.n_chordwise,
                                rake_deg=self.rig.rake_deg, sheer_above_waterline=self.rig.sheer_above_waterline)
        
        geoms = []
        if self.rig.sails_def == 'jib' or self.rig.sails_def == 'jib_and_main':
            jib_geometry = factory.make_jib(
                jib_luff=self.rig.jib_luff,
                foretriangle_base=self.rig.foretriangle_base,
                foretriangle_height=self.rig.foretriangle_height,
                jib_chords=interpolator.interpolate_girths(self.jib_sail.girths, self.jib_sail.chords, self.solver.n_spanwise + 1),
                sail_twist_deg=interpolator.interpolate_girths(self.jib_sail.girths, self.jib_sail.centerline_twist_deg, self.solver.n_spanwise + 1),
                mast_LOA=self.rig.mast_LOA,
                LLT_twist=self.solver.LLT_twist, 
                interpolated_camber=interpolator.interpolate_girths(self.jib_sail.girths, self.jib_sail.camber, self.solver.n_spanwise + 1),
                interpolated_distance_from_luff=interpolator.interpolate_girths(self.jib_sail.girths, self.jib_sail.camber_distance_from_luff, self.solver.n_spanwise + 1)
                )
            geoms.append(jib_geometry)
            
        if self.rig.sails_def == 'main' or self.rig.sails_def =='jib_and_main':
            main_sail_geometry = factory.make_main_sail(
                main_sail_luff=self.rig.main_sail_luff,
                boom_above_sheer=self.rig.boom_above_sheer,
                main_sail_chords=interpolator.interpolate_girths(self.main_sail.girths, self.main_sail.chords, self.solver.n_spanwise + 1),
                sail_twist_deg=interpolator.interpolate_girths(self.main_sail.girths, self.main_sail.centerline_twist_deg, self.solver.n_spanwise + 1),
                LLT_twist=self.solver.LLT_twist,
                interpolated_camber=interpolator.interpolate_girths(self.main_sail.girths, self.main_sail.camber, self.solver.n_spanwise + 1),
                interpolated_distance_from_luff=interpolator.interpolate_girths(self.main_sail.girths, self.main_sail.camber_distance_from_luff, self.solver.n_spanwise + 1)
                )
            geoms.append(main_sail_geometry)

        return SailSet(geoms)
    
    @property
    def sail_set(self):
        return self.__get_sail_set()


import numpy as np
from scipy import interpolate


class Interpolator:
    def __init__(self, interpolation_type):
        self.interpolation_type = interpolation_type

        self.interpolation_switch = {
            "linear": lambda x_new, x, y: np.interp(x_new, x, y),
            "spline": lambda x_new, x, y: interpolate.splev(x_new, interpolate.splrep(x, y, s=0), der=0),
        }

    def interpolate_girths(self, x, y, n_points):
        x_new = np.linspace(0., 1., num=n_points, endpoint=True)
        y_new = self.interpolation_switch[self.interpolation_type](x_new, x, y)
        return y_new

    def interpolate_and_mirror_girths(self, girths, y_data_for_interpolation, n_points):
        chord_sections = self.interpolate_girths(girths, y_data_for_interpolation, n_points)
        chord_sections_mirror = np.flip(chord_sections)
        return np.concatenate([chord_sections_mirror, chord_sections])

    def interpolate_data_for_sections(self, data_main_sail, data_jib, ms_girths, jib_girths, n_span):
        ms_interpolated_sections = [self.interpolate_and_mirror_girths(ms_girths, d, n_span) for d in data_main_sail]
        jib_interpolated_sections = [self.interpolate_and_mirror_girths(jib_girths, d, n_span) for d in data_jib]

        sailset_interpolated_sections = [np.concatenate([j, ms]) for j, ms in
                                         zip(jib_interpolated_sections, ms_interpolated_sections)]
        return sailset_interpolated_sections



import numpy as np
from pySailingVLM.yacht_geometry.sail_geometry import SailGeometry
from pySailingVLM.rotations.csys_transformations import CSYS_transformations


class SailFactory:
    def __init__(self, csys_transformations: CSYS_transformations, n_spanwise=10, n_chordwise=1, rake_deg=90, sheer_above_waterline=0):
        # sin(pi-alpha) = sin(alpha)
        # cos(pi-alpha) = -cos(alpha)

        self._n_spanwise = n_spanwise  # number of panels (span-wise, above the water) per one sail
        self._n_chordwise = n_chordwise  # number of panels (chord-wise, above the water) per one sail
        self.rake = np.deg2rad(rake_deg)
        self.csys_transformations = csys_transformations
        self.sheer_above_waterline_xyz = [-sheer_above_waterline * np.cos(self.rake),
                                          0.,
                                          sheer_above_waterline * np.sin(self.rake)]

    def make_main_sail(self, main_sail_luff, boom_above_sheer, main_sail_chords=None, sail_twist_deg=None, LLT_twist=None, interpolated_camber=None, interpolated_distance_from_luff=None):
        boom_above_sheer_xyz = [-boom_above_sheer * np.cos(self.rake),
                                0.,
                                boom_above_sheer * np.sin(self.rake)]

        # define sail mounting points
        tack_mounting = np.array([self.sheer_above_waterline_xyz[0]+boom_above_sheer_xyz[0],
                                  0.,
                                  self.sheer_above_waterline_xyz[2]+boom_above_sheer_xyz[2]])

        head_mounting = np.array([tack_mounting[0]-main_sail_luff*np.cos(self.rake),
                                  0.,
                                  tack_mounting[2]+main_sail_luff*np.sin(self.rake)])

        main_sail = SailGeometry(head_mounting, tack_mounting, self.csys_transformations,
                                 n_spanwise=self._n_spanwise, n_chordwise=self._n_chordwise,
                                 chords=main_sail_chords,
                                 initial_sail_twist_deg=sail_twist_deg, name="main_sail", LLT_twist=LLT_twist, interpolated_camber=interpolated_camber, interpolated_distance_from_luff=interpolated_distance_from_luff)
        return main_sail

    def make_jib(self, jib_luff, foretriangle_base, foretriangle_height, jib_chords=None, sail_twist_deg=None, mast_LOA=0, LLT_twist=None, interpolated_camber=None, interpolated_distance_from_luff=None):
        # cosine theorem
        forestay_length = np.sqrt(foretriangle_base * foretriangle_base
                                  + foretriangle_height * foretriangle_height
                                  - 2 * foretriangle_base * foretriangle_height * np.cos(self.rake))
        # sine theorem
        forestay_angle = np.arcsin(foretriangle_height * np.sin(self.rake) / forestay_length)

        # define sail mounting points
        tack_mounting = np.array([-foretriangle_base + self.sheer_above_waterline_xyz[0] - mast_LOA/np.sin(self.rake),
                                  0.,
                                  self.sheer_above_waterline_xyz[2]])

        head_mounting = np.array([tack_mounting[0] + jib_luff*np.cos(forestay_angle),
                                  0.,
                                  tack_mounting[2] + jib_luff*np.sin(forestay_angle)])

        jib = SailGeometry(head_mounting, tack_mounting,  self.csys_transformations,
                           n_spanwise=self._n_spanwise, n_chordwise=self._n_chordwise,
                           chords=jib_chords,
                           initial_sail_twist_deg=sail_twist_deg, name="jib", LLT_twist=LLT_twist, interpolated_camber=interpolated_camber, interpolated_distance_from_luff=interpolated_distance_from_luff)
        return jib



import numpy as np
import pandas as pd

from pySailingVLM.solver.forces import calc_moments, calc_moment_arm_in_shifted_csys
from pySailingVLM.solver.forces import determine_vector_from_its_dot_and_cross_product
from pySailingVLM.rotations.csys_transformations import CSYS_transformations
from pySailingVLM.yacht_geometry.sail_geometry import SailSet


from pySailingVLM.solver.additional_functions import extract_above_water_quantities
from pySailingVLM.solver.vlm import Vlm


class InviscidFlowResults:
    def __init__(self,
                 sail_set: SailSet,
                 csys_transformations: CSYS_transformations, myvlm : Vlm):


        self.csys_transformations = csys_transformations
        self.gamma_magnitude = myvlm.gamma_magnitude
        self.pressure = myvlm.pressure
        self.V_induced_at_cp = myvlm.V_induced_at_cp#V_induced
        self.V_induced_length = np.linalg.norm(self.V_induced_at_cp, axis=1)
        self.V_app_fs_at_cp = myvlm.V_app_fs_at_cp#V_app_fs
        self.V_app_fs_length = np.linalg.norm(self.V_app_fs_at_cp, axis=1)
        self.AWA_app_fs = np.arctan(myvlm.V_app_fs_at_cp[:, 1] / myvlm.V_app_fs_at_cp[:, 0])
  
        self.F_xyz = myvlm.force
        self.F_xyz_above_water, self.F_xyz_total = extract_above_water_quantities(self.F_xyz)

        r = calc_moment_arm_in_shifted_csys(myvlm.cp, csys_transformations.v_from_original_xyz_2_reference_csys_xyz)
        ####
        # tests for flat plate
        # remove it later
        # m = np.cross(myvlm.cp, myvlm.force)
        # m_xyz = np.sum(m, axis=0) 
        # m_xyz:
        # array([ 7.16657636e-17, -2.77147486e-16,  1.26635518e+00])
        # matlab:
        # 2.81892564846231e-15, -1.26030647391301, 1.90440111597079e-16
        
        # to jest to samo bo:
        # normals:
        # ...
        # [ 0.,  1., -0.],
        # [ 0.,  1., -0.],
        # [ 0.,  1., -0.],
        # [ 0.,  1., -0.],
        # ...  
        
        # matlab:
        # 0	0	-1
        # 0	0	-1
        # 0	0	-1
        # 0	0	-1
        
        # czyli u nas xyz a w matlabie xzy 
        ####
        dyn_dict = {}
        # for jib and main, quantities is always dividable by 2
        half = int(myvlm.force.shape[0] / 2)
        # dla jiba i maina mamy: jib above, main_above
        # self.F_xyz[0:half] - gorna polowa paneli czyli ta nad woda
        # forces_chunks_above - gorne panele podzielone na chanki, 0 - jib, 1 - main
        forces_chunks_above = np.array_split(self.F_xyz[0:half], len(sail_set.sails))
        r_chunks_above = np.array_split(r[0:half], len(sail_set.sails))

        for i in range(len(sail_set.sails)):

            F_xyz_above_water_tmp = forces_chunks_above[i]
            r_tmp = r_chunks_above[i]

            F_xyz_above_water_tmp_total = np.sum(F_xyz_above_water_tmp, axis=0)
            dyn_dict[f"F_{sail_set.sails[i].name}_total_COG.x"] = F_xyz_above_water_tmp_total[0]
            dyn_dict[f"F_{sail_set.sails[i].name}_total_COG.y"] = F_xyz_above_water_tmp_total[1]
            dyn_dict[f"F_{sail_set.sails[i].name}_total_COG.z"] = F_xyz_above_water_tmp_total[2]

            M_xyz_tmp = calc_moments(r_tmp, F_xyz_above_water_tmp)
            M_xyz_tmp_total = np.sum(M_xyz_tmp, axis=0)
            dyn_dict[f"M_{sail_set.sails[i].name}_total_COG.x"] = M_xyz_tmp_total[0]
            dyn_dict[f"M_{sail_set.sails[i].name}_total_COG.y"] = M_xyz_tmp_total[1]
            dyn_dict[f"M_{sail_set.sails[i].name}_total_COG.z"] = M_xyz_tmp_total[2]

        self.dyn_dict = dyn_dict

        self.M_xyz = calc_moments(r, myvlm.force) 
        _, self.M_total_above_water_in_xyz_csys = extract_above_water_quantities(self.M_xyz)

        r_dot_F = np.array([np.dot(r[i], myvlm.force[i]) for i in range(len(myvlm.force))])
        _, r_dot_F_total_above_water = extract_above_water_quantities(r_dot_F)

        self.above_water_centre_of_effort_estimate_xyz \
            = determine_vector_from_its_dot_and_cross_product(
                self.F_xyz_total, r_dot_F_total_above_water, self.M_total_above_water_in_xyz_csys)

        self.F_centerline = csys_transformations.from_xyz_to_centerline_csys(myvlm.force)
        _, self.F_centerline_total = extract_above_water_quantities(self.F_centerline)

        self.M_centerline_csys = csys_transformations.from_xyz_to_centerline_csys(self.M_xyz)
        _, M_total_above_water_in_centerline_csys = extract_above_water_quantities(self.M_centerline_csys)
        self.M_total_above_water_in_centerline_csys = M_total_above_water_in_centerline_csys

    def estimate_heeling_moment_from_keel(self, underwater_centre_of_effort_xyz):
        self.M_xyz_underwater_estimate_total = np.cross(underwater_centre_of_effort_xyz, -self.F_xyz_total)
        self.M_centerline_csys_underwater_estimate_total = self.csys_transformations.from_xyz_to_centerline_csys([self.M_xyz_underwater_estimate_total])[0]

    def __iter__(self):
        yield 'Circulation_magnitude', self.gamma_magnitude
        yield 'V_induced_COG.x', self.V_induced_at_cp[:, 0]
        yield 'V_induced_COG.y', self.V_induced_at_cp[:, 1]
        yield 'V_induced_COG.z', self.V_induced_at_cp[:, 2]
        yield 'V_induced_length', self.V_induced_length
        yield 'V_app_fs_COG.x', self.V_app_fs_at_cp[:, 0]
        yield 'V_app_fs_COG.y', self.V_app_fs_at_cp[:, 1]
        yield 'V_app_fs_COG.z', self.V_app_fs_at_cp[:, 2]
        yield 'V_app_fs_length', self.V_app_fs_length
        yield 'AWA_fs_COG_deg', np.rad2deg(self.AWA_app_fs)
        yield 'AWA_fs_COW_deg', np.rad2deg(self.AWA_app_fs) - self.csys_transformations.leeway_deg
        yield 'Pressure', self.pressure
        yield 'F_sails_COG.x', self.F_xyz[:, 0]
        yield 'F_sails_COG.y', self.F_xyz[:, 1]
        yield 'F_sails_COG.z', self.F_xyz[:, 2]
        yield 'M_sails_COG.x', self.M_xyz[:, 0]
        yield 'M_sails_COG.y', self.M_xyz[:, 1]
        yield 'M_sails_COG.z', self.M_xyz[:, 2]
        yield 'F_sails_COW.x', self.F_centerline[:, 0]
        yield 'F_sails_COW.y', self.F_centerline[:, 1]
        yield 'F_sails_COW.z', self.F_centerline[:, 2]

    def to_df_full(self):
        obj_as_dict = dict(self)  # this calls __iter__
        df = pd.DataFrame.from_records(obj_as_dict)
        return df

    def to_df_integral(self):
        obj_as_dict = {
            'F_sails_total_COG.x': self.F_xyz_total[0],
            'F_sails_total_COG.y': self.F_xyz_total[1],
            'F_sails_total_COG.z': self.F_xyz_total[2],
            'F_sails_total_COW.x': self.F_centerline_total[0],
            'F_sails_total_COW.y': self.F_centerline_total[1],
            'F_sails_total_COW.z': self.F_centerline_total[2],
            'M_sails_total_COG.x': self.M_total_above_water_in_xyz_csys[0],
            'M_sails_total_COG.y': self.M_total_above_water_in_xyz_csys[1],
            'M_sails_total_COG.z': self.M_total_above_water_in_xyz_csys[2],
            'M_sails_total_COW.x (heel)': self.M_total_above_water_in_centerline_csys[0],
            'M_sails_total_COW.y (pitch)': self.M_total_above_water_in_centerline_csys[1],
            'M_sails_total_COW.z (yaw)': self.M_total_above_water_in_centerline_csys[2],
            'M_sails_total_COW.z (yaw - JG sign)': -self.M_total_above_water_in_centerline_csys[2],
        }

        if hasattr(self, 'M_xyz_underwater_estimate_total'):
            tmp = {
               'M_keel_total_COG.x': self.M_xyz_underwater_estimate_total[0],
                'M_keel_total_COG.y': self.M_xyz_underwater_estimate_total[1],
                'M_keel_total_COG.z': self.M_xyz_underwater_estimate_total[2],
                'M_keel_total_COW.x (heel)': self.M_centerline_csys_underwater_estimate_total[0],
                'M_keel_total_COW.y (pitch)': self.M_centerline_csys_underwater_estimate_total[1],
                'M_keel_total_COW.z (yaw)': self.M_centerline_csys_underwater_estimate_total[2],
                'M_keel_total_COW.z (yaw - JG sign)': -self.M_centerline_csys_underwater_estimate_total[2],
                'M_total_COG.x': self.M_total_above_water_in_xyz_csys[0] + self.M_xyz_underwater_estimate_total[0],
                'M_total_COG.y': self.M_total_above_water_in_xyz_csys[1] + self.M_xyz_underwater_estimate_total[1],
                'M_total_COG.z': self.M_total_above_water_in_xyz_csys[2] + self.M_xyz_underwater_estimate_total[2],
                'M_total_COW.x (heel)': self.M_total_above_water_in_centerline_csys[0] +
                                        self.M_centerline_csys_underwater_estimate_total[0],
                'M_total_COW.y (pitch)': self.M_total_above_water_in_centerline_csys[1] +
                                         self.M_centerline_csys_underwater_estimate_total[1],
                'M_total_COW.z (yaw)': self.M_total_above_water_in_centerline_csys[2] +
                                       self.M_centerline_csys_underwater_estimate_total[2],
                'M_total_COW.z (yaw - JG sign)': -(self.M_total_above_water_in_centerline_csys[2] +
                                                   self.M_centerline_csys_underwater_estimate_total[2]),
           
            }
            obj_as_dict.update(tmp)

        obj_as_dict.update(self.dyn_dict)
        df = pd.DataFrame.from_records(obj_as_dict, index=['Value']).transpose()
        df.reset_index(inplace=True)  # Convert the index (0-th like column) to 'regular' Column
        df = df.rename(columns={'index': 'Quantity'})
        return df



import timeit
import numpy as np


from pySailingVLM.rotations.geometry_calc import rotation_matrix
from pySailingVLM.solver.panels import make_panels_from_le_te_points, get_panels_area
from pySailingVLM.solver.coefs import get_CL_CD_free_wing, get_vlm_CL_CD_free_wing, get_vlm_Cxyz
from pySailingVLM.solver.coefs import calculate_normals_collocations_cps_rings_spans_leading_trailing_mid_points, \
                                solve_eq, calculate_RHS, calc_velocity_coefs

from pySailingVLM.solver.velocity import calculate_app_fs 
from pySailingVLM.solver.forces import is_no_flux_BC_satisfied, calc_force_wrapper, calc_pressure

class Aircraft:
    def __init__(self, chord_length : float, half_wing_span: float, AoA_deg : float, n_spanwise : int, n_chordwise, V : np.array, rho : float = 1.225, gamma_orientation = 1.0) -> None:
        self.chord_length = chord_length
        self.half_wing_span = half_wing_span
        self.AoA_deg = AoA_deg
        self.n_spanwise = n_spanwise
        self.n_chordwise = n_chordwise
        self.V = V
        self.rho = rho
        self.gamma_orientation = gamma_orientation

        
        
        self.__AR = 2 * self.half_wing_span / self.chord_length # aspect ratio
        self.__S = 2 * self.half_wing_span * self.chord_length
        
        self.__CL_theoretical, self.__CD_theoretical = get_CL_CD_free_wing(self.__AR, self.AoA_deg)

        
        # le_NW = np.array([0., self.half_wing_span, 0.])      # leading edge North - West coordinate
        # le_SW = np.array([0., -self.half_wing_span, 0.])     # leading edge South - West coordinate

        # te_NE = np.array([self.chord_length, self.half_wing_span, 0.])   # trailing edge North - East coordinate
        # te_SE = np.array([self.chord_length, -self.half_wing_span, 0.])  # trailing edge South - East coordinate
        
        # Ry = rotation_matrix([0, 1, 0], np.deg2rad(self.AoA_deg))
        
        le_NW = np.array([0., 0 , self.half_wing_span])      # leading edge North - West coordinate
        le_SW = np.array([0., 0., -self.half_wing_span])     # leading edge South - West coordinate

        te_NE = np.array([self.chord_length, 0., self.half_wing_span])   # trailing edge North - East coordinate
        te_SE = np.array([self.chord_length, 0., -self.half_wing_span])  # trailing edge South - East coordinate
        
        Rz = rotation_matrix([0, 0, 1], np.deg2rad(-self.AoA_deg))

        panels, trailing_edge_info, leading_edge_info = make_panels_from_le_te_points([np.dot(Rz, le_SW),
                                                                                    np.dot(Rz, te_SE),
                                                                                    np.dot(Rz, le_NW),
                                                                                    np.dot(Rz, te_NE)],
                                                                                    [self.n_chordwise, self.n_spanwise])
        
        N = panels.shape[0]
        self.__V_app_infw = np.array([self.V for i in range(N)])
        
        areas = get_panels_area(panels)
        normals, ctr_p, cp, rings, span_vectors, _, _ = calculate_normals_collocations_cps_rings_spans_leading_trailing_mid_points(panels, self.gamma_orientation)
        coefs, v_ind_coeff = calc_velocity_coefs(self.__V_app_infw, ctr_p, rings, normals, trailing_edge_info, self.gamma_orientation)
        self.__RHS = calculate_RHS(self.__V_app_infw, normals)
        self.__gamma_magnitude = solve_eq(coefs, self.__RHS)
        
        _,  V_app_fs_at_ctrl_p = calculate_app_fs(self.__V_app_infw,  v_ind_coeff,  self.__gamma_magnitude)
        assert is_no_flux_BC_satisfied(V_app_fs_at_ctrl_p, panels, areas, normals)
        self.__force, _, _ = calc_force_wrapper(self.__V_app_infw, self.__gamma_magnitude, self.rho, cp, rings, self.n_spanwise, normals, span_vectors, trailing_edge_info, leading_edge_info, self.gamma_orientation)
        self.__pressure = calc_pressure(self.__force, normals, areas)

        self.__CL, self.__CD = get_vlm_CL_CD_free_wing(self.__force, self.V, self.rho, self.__S)

        total_p = np.sum(self.__pressure, axis=0)
        q = 0.5 * rho * (np.linalg.norm(self.V) ** 2) * self.__S
        self.__Cp = total_p / q

    def get_info(self):
        print("gamma_magnitude: \n")
        print(self.__gamma_magnitude)
        print("DONE")
        print(f"\nAspect Ratio {self.__AR}")
        print(f"CL_expected {self.__CL_theoretical:.6f} \t CD_ind_expected {self.__CD_theoretical:.6f}")
        print(f"CL_vlm      {self.__CL:.6f}  \t CD_vlm          {self.__CD:.6f}")
        print(f"\n\ntotal_F {str(np.sum(self.__force, axis=0))}")

    def __get_Cxyz(self):
        return get_vlm_Cxyz(self.__force, self.V, self.rho, self.__S)

    @property
    def Cxyz(self):
        return self.__get_Cxyz()
    
    @property
    def CL_theoretical(self):
        return self.__CL_theoretical

    @property
    def CD_theoretical(self):
        return self.__CD_theoretical
    
    @property
    def AR(self):
        return self.__AR

    @property
    def CL(self):
        return self.__CL

    @property
    def CD(self):
        return self.__CD
    @property
    def Cp(self):
        return self.__Cp
    @property
    def force(self):
        return self.__force

    @property
    def pressure(self):
        return self.__pressure









from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pySailingVLM.solver.vlm import Vlm
    from pySailingVLM.results.inviscid_flow import InviscidFlowResults
    from pySailingVLM.inlet.inlet_conditions import InletConditions
    from pySailingVLM.yacht_geometry.hull_geometry import HullGeometry
import numpy as np
import math

import matplotlib.pyplot as plt

from pySailingVLM.solver.additional_functions import normalize
from matplotlib.patches import FancyArrowPatch
from mpl_toolkits.mplot3d import proj3d
import mpl_toolkits.mplot3d as a3
import matplotlib.cm as cm
import matplotlib as mpl

from typing import Tuple

import spatialpandas as sp
import holoviews as hv
from holoviews.streams import PlotSize

class Arrow3D(FancyArrowPatch):
    def __init__(self, xs, ys, zs, *args, **kwargs):
        FancyArrowPatch.__init__(self, (0, 0), (0, 0), *args, **kwargs)
        self._verts3d = xs, ys, zs

    def do_3d_projection(self, renderer=None):
        xs3d, ys3d, zs3d = self._verts3d
        xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, self.axes.M)
        self.set_positions((xs[0],ys[0]),(xs[1],ys[1]))

        return np.min(zs)

    def draw(self, renderer):
        xs3d, ys3d, zs3d = self._verts3d
        xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, self.axes.M)
        self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))
        FancyArrowPatch.draw(self, renderer)


def display_panels_xyz(vlm : Vlm, color_panels_by : str) -> Tuple[plt.axes, int]:
    """
    display_panels_xyz display panels in 3 dimentions

    :param Vlm vlm: vlm class instance
    :param color_panels_by: gamma_magnitude or p_coeffs string
    :return List[plt.axes, int]: list containing matplotlib plt ax and size of water (int)
    """
    fig = plt.figure(figsize=(12, 12))
    ax = plt.axes(projection='3d')

    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')

    def set_ax_eq(ax : plt.Axes, X : np.ndarray, Y : np.ndarray, Z : np.ndarray) -> float:
        """
        set_ax_eq sets limits on x, y, z axes

        :param plt.Axes ax: ax to be modified
        :param np.ndarray X: x array
        :param np.ndarray Y: y array
        :param np.ndarray Z: z array
        :return float: max range number 
        """
        # ax.set_aspect('equal') - matplotlib bug
        # dirty hack: NotImplementedError: It is not currently possible to manually set the aspect on 3D axes
        max_range = np.array([X.max() - X.min(), Y.max() - Y.min(), Z.max() - Z.min()]).max() / 2.0

        mid_x = (X.max() + X.min()) * 0.5
        mid_y = (Y.max() + Y.min()) * 0.5
        mid_z = (Z.max() + Z.min()) * 0.5
        ax.set_xlim(mid_x - max_range, mid_x + max_range)
        ax.set_ylim(mid_y - max_range, mid_y + max_range)
        ax.set_zlim(mid_z - max_range, mid_z + max_range)

        return max_range

    data = None
    if color_panels_by == 'gamma_magnitude':
        data = vlm.pressure
    elif color_panels_by == 'p_coeffs':
        data = vlm.p_coeffs
    else:
        data = vlm.pressure
    norm = mpl.colors.Normalize(vmin=min(data), vmax=max(data))
    cmap = cm.hot
    m = cm.ScalarMappable(norm=norm, cmap=cmap)

    for vtx, d in zip(vlm.panels, data):
        tri = a3.art3d.Poly3DCollection([vtx])
        tri.set_color(m.to_rgba(d))
        tri.set_edgecolor('k')
        ax.add_collection3d(tri)

    fig.colorbar(m, ax=ax)
    ###
    # plot water level
    water_size = int(1.1*math.ceil(max(max(abs(vlm.leading_mid_points[:, 2])), max(abs(vlm.trailing_mid_points[:, 2])))))
    xx, yy = np.meshgrid(range(-water_size, water_size), range(-water_size, water_size))
    zz = 0 * xx + 0 * yy
    ax.plot_surface(xx, yy, zz, alpha=0.25)

    set_ax_eq(ax, xx, yy, zz)
    return ax, water_size

def display_panels_or_rings(things : np.ndarray, pressure : np.ndarray, leading_mid_points : np.ndarray, trailing_mid_points : np.ndarray) -> Tuple[plt.axes, int]:
    """
    display_panels_or_rings display panels or rings  in 3 dimentions

    :param Vlm vlm: vlm class instance
    :return List[plt.axes, int]: list containing matplotlib plt ax and size of water (int)
    """
    fig = plt.figure(figsize=(12, 12))
    ax = plt.axes(projection='3d')

    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')

    def set_ax_eq(ax : plt.Axes, X : np.ndarray, Y : np.ndarray, Z : np.ndarray) -> float:
        """
        set_ax_eq sets limits on x, y, z axes

        :param plt.Axes ax: ax to be modified
        :param np.ndarray X: x array
        :param np.ndarray Y: y array
        :param np.ndarray Z: z array
        :return float: max range number 
        """
        # ax.set_aspect('equal') - matplotlib bug
        # dirty hack: NotImplementedError: It is not currently possible to manually set the aspect on 3D axes
        max_range = np.array([X.max() - X.min(), Y.max() - Y.min(), Z.max() - Z.min()]).max() / 2.0

        mid_x = (X.max() + X.min()) * 0.5
        mid_y = (Y.max() + Y.min()) * 0.5
        mid_z = (Z.max() + Z.min()) * 0.5
        ax.set_xlim(mid_x - max_range, mid_x + max_range)
        ax.set_ylim(mid_y - max_range, mid_y + max_range)
        ax.set_zlim(mid_z - max_range, mid_z + max_range)

        return max_range


    #norm = mpl.colors.Normalize(vmin=min(pressure), vmax=max(pressure))
    #cmap = cm.hot
    #m = cm.ScalarMappable(norm=norm, cmap=cmap)

    for vtx, p in zip(things, pressure):
        tri = a3.art3d.Poly3DCollection([vtx])
        #tri.set_color(m.to_rgba(p))
        tri.set_edgecolor('k')
        ax.add_collection3d(tri)

    #fig.colorbar(m, ax=ax)
    ###
    # plot water level
    water_size = int(1.1*math.ceil(max(max(abs(leading_mid_points[:, 2])), max(abs(trailing_mid_points[:, 2])))))
    xx, yy = np.meshgrid(range(-water_size, water_size), range(-water_size, water_size))
    zz = 0 * xx + 0 * yy
    ax.plot_surface(xx, yy, zz, alpha=0.25)

    set_ax_eq(ax, xx, yy, zz)
    return ax, water_size

def display_hull(ax: plt.Axes, hull: HullGeometry):
    """
    display_hull plots hull 

    :param plt.Axes ax: axes for drawing
    :param HullGeometry hull: hull object
    """
    ax.plot(hull.deck_centerline[:, 0], hull.deck_centerline[:, 1], hull.deck_centerline[:, 2], 'gray')
    ax.plot(hull.deck_port_line[:, 0], hull.deck_port_line[:, 1], hull.deck_port_line[:, 2], 'gray')
    ax.plot(hull.deck_starboard_line[:, 0], hull.deck_starboard_line[:, 1], hull.deck_starboard_line[:, 2], 'gray')
    ax.plot(hull.deck_port_line_underwater[:, 0], hull.deck_port_line_underwater[:, 1], hull.deck_port_line_underwater[:, 2], 'gray', alpha=0.25)
    ax.plot(hull.deck_starboard_line_underwater[:, 0], hull.deck_starboard_line_underwater[:, 1], hull.deck_starboard_line_underwater[:, 2], 'gray', alpha=0.25)


def display_winds(ax : plt.Axes, cp_points : np.ndarray, water_size : int,  inlet_condition: InletConditions, inviscid_flow_results :  InviscidFlowResults, n_spanwise : int, n_chordwise : int, show_induced_wind: bool = False):
    """
    display_winds displays winds on final plot

    :param plt.Axes ax: ax
    :param np.ndarray cp_points: array with center of pressure points
    :param int water_size: size of water
    :param InletConditions inlet_condition: inlet conditions
    :param InviscidFlowResults inviscid_flow_results: flow results
    """
    #N = len(cp_points[:, 2])
     
    mean_AWA = np.mean(inlet_condition.AWA_infs_deg)
    shift_x = (-0.925) * water_size * np.cos(np.deg2rad(mean_AWA))
    shift_y = (-0.925) * water_size * np.sin(np.deg2rad(mean_AWA))

    #######
    # example:
    # cp_points has 24 points, n_spanwose = 2, n_chordise=3 it means that
    # we have: 24 / (2*3) = 24 / 6 = 4 elements: jib + main above water and jib and main under water
    l = int(cp_points.shape[0] / (n_spanwise * n_chordwise))
    # if you have main and jib this list should have 4 elements
    # if only jib - 2 elements
    
    ###
    #####
    # get winds acting only on first spanwise section on each sail
    tws_chunks = np.split(inlet_condition.tws_at_cp, n_chordwise*l)
    tws_chunks = np.asarray(tws_chunks)
    tws_chunks_at_first_spanwise_sections = tws_chunks[0::n_chordwise]
    sh0, sh1, _ = tws_chunks_at_first_spanwise_sections.shape
    tws_chunks_at_first_spanwise_sections = tws_chunks_at_first_spanwise_sections.reshape(sh0 * sh1 ,3)
    
    V_app_infs_chunks = np.split(inlet_condition.V_app_infs, n_chordwise*l)
    V_app_infs_chunks = np.asarray(V_app_infs_chunks)
    V_app_infs_at_first_spanwise_sections = V_app_infs_chunks[0::n_chordwise]
    sh0, sh1, _ = V_app_infs_at_first_spanwise_sections.shape
    V_app_infs_at_first_spanwise_sections = V_app_infs_at_first_spanwise_sections.reshape(sh0 * sh1 ,3)
    
    cp_points_chunks = np.split(inlet_condition.cp_points, n_chordwise*l)
    cp_points_chunks = np.asarray(cp_points_chunks)
    cp_points_at_first_spanwise_sections = cp_points_chunks[0::n_chordwise]
    sh0, sh1, _ = cp_points_at_first_spanwise_sections.shape
    cp_points_at_first_spanwise_sections = cp_points_at_first_spanwise_sections.reshape(sh0 * sh1 ,3)
    
    # assert
    V_winds = [tws_chunks_at_first_spanwise_sections, V_app_infs_at_first_spanwise_sections]
    colors = ['green', 'blue']  # G: True wind, B: - Apparent wind, R: Induced@CP
    if show_induced_wind:
        V_induced_at_cp_chunks = np.split(inviscid_flow_results.V_induced_at_cp, n_chordwise*l)
        V_induced_at_cp_chunks = np.asarray(V_induced_at_cp_chunks)
        V_induced_at_cp_chunks_at_first_spanwise_sections = V_induced_at_cp_chunks[0::n_chordwise]
        sh0, sh1, _ = V_induced_at_cp_chunks_at_first_spanwise_sections.shape
        V_induced_at_cp_chunks_at_first_spanwise_sections = V_induced_at_cp_chunks_at_first_spanwise_sections.reshape(sh0 * sh1 ,3)
        V_winds.append(V_induced_at_cp_chunks_at_first_spanwise_sections)
        colors.append('red')

    for arr in V_winds:
        assert arr.shape[0] == n_spanwise * l, 'Not proper array shape, check lines above!'

    app_induced_colors = ['peru', 'red', 'peru', 'red']
    if l == 2:
        app_induced_colors = ['peru', 'red']
    # check if colors are defined for appaernt + induced wind - above and under water
    assert (len(app_induced_colors) == l and len(app_induced_colors) % 2 == 0), "Bad length of app_induced_colors list"
    color_counter = 0
    for V_wind, color in zip(V_winds, colors):
        for i in range(n_spanwise*l):
            # for chnaging colors for induced wind based on sail (jib above, main above, jib_underwater, main_underwater etc) 
            draw_color = color
            if color == 'red':
                if i % n_spanwise == 0 and i !=0:
                    color_counter += 1
                draw_color = app_induced_colors[color_counter]
                    
                shift_x = cp_points_at_first_spanwise_sections[i, 0]
                shift_y = cp_points_at_first_spanwise_sections[i, 1]

            vx = np.array([shift_x, shift_x+V_wind[i, 0]])
            vy = np.array([shift_y, shift_y+V_wind[i, 1]])
            vz = np.array([cp_points_at_first_spanwise_sections[i, 2], cp_points_at_first_spanwise_sections[i, 2]])

            if cp_points_at_first_spanwise_sections[i, 2] > 0:
                arrow = Arrow3D(vx, vy, vz, mutation_scale=10, lw=1, arrowstyle="-|>", color=draw_color, alpha=0.75)
            else:
                arrow = Arrow3D(vx, vy, vz, mutation_scale=10, lw=1, arrowstyle="-|>", color=draw_color, alpha=0.05)
            ax.add_artist(arrow)


def display_CE_CLR(ax : plt.Axes,
                   inviscid_flow_results: InviscidFlowResults,
                   hull: HullGeometry) -> Tuple[float, np.ndarray, np.ndarray]:
    """
    display_CE_CLR display force 

    :param plt.Axes ax: plot ax
    :param InviscidFlowResults inviscid_flow_results: flow results
    :param HullGeometry hull: hull object
    :return Tuple[float, np.ndarray, np.ndarray]: Tuple[scale, lateral resistance, effort, force]
    """

    # https://en.wikipedia.org/wiki/Forces_on_sails#Forces_on_sailing_craft
    def plot_vector(origin, length):
        vx = np.array([origin[0], origin[0] + length[0]])
        vy = np.array([origin[1], origin[1] + length[1]])
        vz = np.array([origin[2], origin[2] + length[2]])
        arrow = Arrow3D(vx, vy, vz, mutation_scale=10, lw=1, arrowstyle="-|>", color='black', alpha=0.75)
        ax.add_artist(arrow)
        ax.scatter3D(origin[0], origin[1], origin[2], c='black', marker="o")

    scale = 0.8*np.mean(inviscid_flow_results.V_app_fs_length)  # ~10 gives nice plot
    clr = hull.center_of_lateral_resistance
    ce = inviscid_flow_results.above_water_centre_of_effort_estimate_xyz

    F = scale * normalize(inviscid_flow_results.F_xyz_total)
    plot_vector(ce, F)
    plot_vector(clr, -F)
    
    return scale, clr, ce, F

def display_panels_xyz_and_winds(vlm :Vlm, inviscid_flow_results: InviscidFlowResults, 
                                 inlet_condition: InletConditions,
                                 hull: HullGeometry,
                                 show_plot: bool = True, show_induced_wind: bool = False,
                                 is_sailopt_mode=False
                                 ):
    """
    display_panels_xyz_and_winds plot whole yacht with winds and force

    :param Vlm vlm: Vlm object
    :param InviscidFlowResults inviscid_flow_results: flow results
    :param InletConditions inlet_condition: inlet conditions
    :param HullGeometry hull: hull object
    :param bool show_plot: decide if plot should be shown, defaults to True
    """
    
    color_panels_by = None
    title_mode = None
    if is_sailopt_mode:
        color_panels_by = 'gamma_magnitude'
        title_mode = 'SailOpt mode: Panels colored by optimal circulation \n'
    else:
        color_panels_by = 'p_coeffs'
        title_mode = 'VLM mode: Panels colored by coefficient of pressure \n'


    ax, water_size = display_panels_xyz(vlm, color_panels_by)

    red_text = ''
    if show_induced_wind:
        red_text = ', Induced@CP (red)' 
    
    ax.set_title(f'{title_mode}'
                 f'Winds: True (green), Apparent (blue){red_text} \n'
                 'Centre of Effort & Center of Lateral Resistance (black)')
    
    
    display_hull(ax, hull)

    display_winds(ax, vlm.cp, water_size, inlet_condition, inviscid_flow_results, vlm.n_spanwise, vlm.n_chordwise, show_induced_wind)

    scale, clr, ce, F = display_CE_CLR(ax, inviscid_flow_results, hull)
    

    if show_plot:
        plt.show()


def plot_cp(mesh : np.ndarray, p_coeffs : np.ndarray, path_to_save : str):
    hv.renderer('bokeh')
    boundaries = sp.geometry.PolygonArray([[np.vstack([panel[:,[0,2]], panel[:,[0,2]][0]]).flatten()] for panel in mesh])
    info = sp.GeoDataFrame({'boundary': boundaries, 'p_coeffs': list(p_coeffs)}) 
    
    ropts = dict(tools=["hover"], height=380, width=330, colorbar=True, colorbar_position="right", color='p_coeffs')
    hvpolys = hv.Polygons(info, vdims=['p_coeffs']).opts(**ropts)
    
    try:
        __IPYTHON__
        from bokeh.io import show, output_notebook
        output_notebook()
        show(hv.render(hvpolys, backend='bokeh'))
    except NameError:
        hv.save(hvpolys, path_to_save + '/cp_plot.html', backend='bokeh')


import numpy as np
import matplotlib.pyplot as plt 

from abc import abstractmethod, ABC
from pySailingVLM.rotations.geometry_calc import rotate_points_around_arbitrary_axis

from pySailingVLM.rotations.csys_transformations import CSYS_transformations

from typing import List

from pySailingVLM.solver.panels import make_panels_from_mesh_spanwise
from pySailingVLM.solver.mesher import make_airfoil_mesh
#from pySailingVLM.solver.additional_functions import plot_mesh
class BaseGeometry:

    @property
    @abstractmethod
    def panels(self):
        pass

    
class SailGeometry(BaseGeometry, ABC):
    def __init__(self, head_mounting: np.ndarray, tack_mounting: np.ndarray,
                 csys_transformations: CSYS_transformations,
                 n_spanwise=10, n_chordwise=1, chords=None,
                 initial_sail_twist_deg=None, name=None, LLT_twist=None,  interpolated_camber=None, interpolated_distance_from_luff=None
                 ):

        self.__n_spanwise = n_spanwise  # number of panels (span-wise) - above the water
        self.__n_chordwise = n_chordwise  # number of panels (chord-wise) - in LLT there is line instead of panels
        self.name = name

        self.csys_transformations = csys_transformations

        self.head_mounting = head_mounting
        self.tack_mounting = tack_mounting
        """
            The geometry is described using the following CSYS.
            le - leading edge (luff) of the sail
            te - trailing edge (leech) of the sail
            below is example for the main sail.
            same for jib.

                        Z ^ (mast)     
                          |
                         /|
                        / |
                       /  |              ^ Y     
                   lem_NW +--+tem_NE    / 
                     /    |   \        /
                    /     |    \      /
                   /      |     \    /
                  /       |      \  /
                 /        |       \/
                /         |       /\
               /          |      /  \
              /           |     /    \
             /            |    /      \
            /      lem_SW |---/--------+tem_SE
           /              |  /          
  (bow) ------------------|-/-------------------------| (stern)
         \                |/                          |
    ------\---------------*---------------------------|-------------------------> X (water level)

        """

        le_NW = head_mounting
        le_SW = tack_mounting

        # mirror z coord in water surface
        # remember that direction of the lifting-line matters
        le_NW_underwater = np.array(
            [tack_mounting[0], tack_mounting[1], -tack_mounting[2]])  # leading edge South - West coordinate - mirror
        le_SW_underwater = np.array(
            [head_mounting[0], head_mounting[1], -head_mounting[2]])  # leading edge North - West coordinate - mirror
        
        
        chords_vec = np.array([chords, np.zeros(len(chords)), np.zeros(len(chords))])
        chords_vec = chords_vec.transpose()
        fchords_vec = np.flip(chords_vec, axis=0)
        
        #### state "zero"
        mesh = make_airfoil_mesh([le_SW, le_NW],[self.__n_chordwise, self.__n_spanwise],chords_vec, interpolated_distance_from_luff, interpolated_camber)
        ###
        zero_mesh = np.swapaxes(mesh, 0, 1)
        panels_above_zero, _, _= make_panels_from_mesh_spanwise(zero_mesh)
        ###
        
        sh0, sh1, sh2 = mesh.shape
        mesh = mesh.reshape(sh0*sh1, sh2)
        
        mesh_underwater = make_airfoil_mesh([le_SW_underwater, le_NW_underwater],[self.__n_chordwise, self.__n_spanwise],fchords_vec, interpolated_distance_from_luff, interpolated_camber)
        ### zero mesh
        zero_mesh_under = np.swapaxes(mesh_underwater, 0, 1)
        panels_under_zero, _, _= make_panels_from_mesh_spanwise(zero_mesh_under)
        ### 
        mesh_underwater = mesh_underwater.reshape(sh0*sh1, sh2)
    
        self.panels_under_zero = panels_under_zero
        self. panels_above_zero = panels_above_zero
        ### end of zero mesh
        
        # rotation # for heel
        #mesh:  le NW p1 p2 ... te
        #       le p1 p2 ... te
        #       le p1 p2 ... te
        #       le SW p2 ... te
        rmesh = np.array([self.csys_transformations.rotate_point_with_mirror(point) for point in mesh])
        rmesh_underwater = np.array([self.csys_transformations.rotate_point_with_mirror(point) for point in mesh_underwater])

        mesh = rmesh
        mesh_underwater = rmesh_underwater
        
        ## twist
        if initial_sail_twist_deg is not None and LLT_twist is not None:
            le_NW = self.csys_transformations.rotate_point_with_mirror(le_NW)
            le_SW = self.csys_transformations.rotate_point_with_mirror(le_SW)
            le_SW_underwater = self.csys_transformations.rotate_point_with_mirror(le_SW_underwater)
            le_NW_underwater = self.csys_transformations.rotate_point_with_mirror(le_NW_underwater)
        
            # print(f"Applying initial_sail_twist_deg to {self.name} -  Lifting Line, mode: {LLT_twist}")
            twist_dict = {
                'sheeting_angle_const': np.full(len(initial_sail_twist_deg), np.min(initial_sail_twist_deg)),
                'average_const': np.full(len(initial_sail_twist_deg), np.average(initial_sail_twist_deg)),
                'real_twist': initial_sail_twist_deg
            }
            sail_twist_deg = twist_dict[LLT_twist]
            sail_twist_deg = np.hstack([initial_sail_twist_deg] * (sh1))
            sail_twist_deg = sail_twist_deg.reshape(sh1, sh0).transpose().flatten()

            p2 = mesh[::sh1][-1]
            p1 = mesh[::sh1][0]
            trmesh = self.rotate_points_around_le(mesh, p1, p2, sail_twist_deg)
            # check if points on forestay are not rotated (they are on axis of rotation)
            np.testing.assert_almost_equal(trmesh[::sh1], mesh[::sh1])
           
            p2_u = mesh_underwater[::sh1][-1]
            p1_u = mesh_underwater[::sh1][0]
            trmesh_underwater = self.rotate_points_around_le(mesh_underwater, p1_u, p2_u, np.flip(sail_twist_deg))
            # check if points on forestay are not rotated (they are on axis of rotation)
            np.testing.assert_almost_equal(trmesh_underwater[::sh1], mesh_underwater[::sh1])

            mesh = trmesh
            mesh_underwater = trmesh_underwater
     
        # come back to original shape 
        mesh = mesh.reshape(sh0, sh1, sh2)
        mesh = np.swapaxes(mesh, 0, 1)
        mesh_underwater = mesh_underwater.reshape(sh0, sh1, sh2)
        mesh_underwater = np.swapaxes(mesh_underwater, 0, 1)
        
        new_approach_panels, trailing_edge_info, leading_edge_info= make_panels_from_mesh_spanwise(mesh)
        new_approach_panels_mirror, trailing_edge_info_mirror, leading_edge_info_mirror = make_panels_from_mesh_spanwise(mesh_underwater)
        
        np.testing.assert_array_equal(trailing_edge_info, trailing_edge_info_mirror)
        np.testing.assert_array_equal(leading_edge_info, leading_edge_info_mirror)
        
        self.__panels_above = new_approach_panels
        self.__panels_under = new_approach_panels_mirror
        self.__panels = np.concatenate([self.__panels_above, self.__panels_under])
        
        # both trailing_edge_info and leading_edge_info are the same for above and underwater
        self.__trailing_edge_info = trailing_edge_info
        self.__leading_edge_info = leading_edge_info

        
        
    def rotate_points_around_le(self, points, p1, p2, sail_twist_deg_vec):
     
        # if all elements are the same -> only once do the calculation -> less matrix multiplications
        result = np.all(sail_twist_deg_vec == sail_twist_deg_vec[0])
        rotated_points = points
        if result:
            rotated_points = rotate_points_around_arbitrary_axis(points, p1, p2, np.deg2rad(sail_twist_deg_vec[0]))
        else:
            # squezee removes "1" diemntion
            # rotate_points_around_arbitrary_axis needs [[a, b, c]] or [[a, b, c], [e, f, g], ...] 
            rotated_points = np.array([np.squeeze(rotate_points_around_arbitrary_axis(np.array([point]), p1, p2, np.deg2rad(deg))) for point, deg in zip(points, sail_twist_deg_vec) ])

        return rotated_points
    
    @property
    def n_chordwise(self):
        return self.__n_chordwise
    
    @property
    def n_spanwise(self):
        return self.__n_spanwise
    
    @property
    def panels(self):
        return self.__panels
    
    @property
    def panels_above(self):
        return self.__panels_above
    
    @property
    def panels_under(self):
        return self.__panels_under
    
    @property
    def trailing_edge_info(self):
        return self.__trailing_edge_info
    
    @property
    def leading_edge_info(self):
        return self.__leading_edge_info
    

class SailSet(BaseGeometry):
    def __init__(self, sails: List[SailGeometry]):
        self.sails = sails
    
        panels_above = np.concatenate([sail.panels_above for sail in self.sails])
        panels_under = np.concatenate([sail.panels_under for sail in self.sails])
        self.__panels = np.concatenate([panels_above, panels_under])
        
        trailing_info_above = np.concatenate([sail.trailing_edge_info for sail in self.sails])
        trailing_info_under = np.concatenate([sail.trailing_edge_info for sail in self.sails])
        self.__trailing_edge_info = np.concatenate([trailing_info_above, trailing_info_under])
        
        leading_info_above = np.concatenate([sail.leading_edge_info for sail in self.sails])
        leading_info_under = np.concatenate([sail.leading_edge_info for sail in self.sails])
        self.__leading_edge_info = np.concatenate([leading_info_above, leading_info_under])

        # mesh without any rotation (in 2d, one cooridinate is 0)
        panels_above_zero = np.concatenate([sail.panels_above_zero for sail in self.sails])
        panels_under_zero = np.concatenate([sail.panels_under_zero for sail in self.sails])
        self.zero_mesh = np.concatenate([panels_above_zero, panels_under_zero])
    @property
    def panels(self):
        return self.__panels
    
    @property
    def trailing_edge_info(self):
        return self.__trailing_edge_info
    
    @property
    def leading_edge_info(self):
        return self.__leading_edge_info



# DO NOT DELETE ME !!!
# USED BY TYPE CHECKING IN DATACLASSES
import os
import numpy as np
import time
from dataclasses import dataclass, field

# OUTPUT DIR
@dataclass
class Output:
    case_name: str = os.path.basename(__file__)  # get name of the current file
    case_dir: str = os.path.dirname(os.path.realpath(__file__)) # get dir of the current file
    name: str = os.path.join("results_example_jib_and_mainsail_vlm", time.strftime("%Y-%m-%d_%Hh%Mm%Ss"))
    file_name: str = 'my_fancy_results' # name of xlsx excel file

# SOLVER SETTINGS
@dataclass
class Solver:
    n_spanwise: int  = 4  # No of control points (above the water) per sail, recommended: 50
    n_chordwise: int = 4 # No of control points (above the water) per sail, recommended: 50
    interpolation_type: str = "spline"  # either "spline" or "linear"
    LLT_twist: str = "real_twist"  # defines how the Lifting Line discretize the sail twist.
    # It can be "sheeting_angle_const" or "average_const" or "real_twist"


# SAILING CONDITIONS
@dataclass
class Conditions:
    leeway_deg: float = 5.    # [deg]
    heel_deg: float = 10.     # [deg]
    SOG_yacht: float = 4.63   # [m/s] yacht speed - speed over ground (leeway is a separate variable)
    tws_ref: float = 4.63     # [m/s] true wind speed
    alpha_true_wind_deg: float = 50.   # [deg] true wind angle (with reference to course over ground) => Course Wind Angle to the boat track = true wind angle to centerline + Leeway
    reference_water_level_for_wind_profile: float = -0.  # [m] this is an attempt to mimick the deck effect
    # by lowering the sheer_above_waterline
    # while keeping the wind profile as in original geometry
    # this shall be negative (H = sail_ctrl_point - water_level)
    wind_exp_coeff: float = 0.1428  # [-] coefficient to determine the exponential wind profile
    wind_reference_measurment_height: float = 10.  # [m] reference height for exponential wind profile
    rho: float = 1.225 # air density [kg/m3]
    wind_profile: str = 'exponential'# allowed: 'exponential' or 'flat' or 'logarithmic'
    roughness: float = 0.05 # for logarithmic profile only 



# GEOMETRY OF THE RIG
@dataclass
class Rig:
    main_sail_luff: float = 12.4  # [m]
    jib_luff: float = 10.0  # [m] 
    foretriangle_height: float = 11.50  # [m]
    foretriangle_base: float = 3.90  # [m]
    sheer_above_waterline: float = 1.2 #[m]
    boom_above_sheer: float = 1.3 # [m]
    rake_deg: float = 92.  # rake angle [deg]
    mast_LOA: float = 0.15  # [m]
    sails_def: float = 'jib_and_main' # definition of sail set, possible: 'jib' or 'main' or 'jib_and_main'


# INPUT - GEOMETRY OF THE SAIL
# INFO for camber:
# First digit describing maximum camber as percentage of the chord.
# Second digit describing the distance of maximum camber from the airfoil leading edge in tenths of the chord.

@dataclass
class MainSail:
    centerline_twist_deg: np.ndarray = None
    
    girths: np.ndarray = np.array([0.00, 1./8, 1./4, 1./2, 3./4, 7./8, 1.00])
    chords: np.ndarray = np.array([4.00, 3.82, 3.64, 3.20, 2.64, 2.32, 2.00])
    camber: np.ndarray = 5*np.array([0.01, 0.01, 0.01, 0.01, 0.01, 0.01, 0.01])
    camber_distance_from_luff: np.ndarray = np.array([0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5]) # starting from leading edge
    
    def __post_init__(self):
        if self.centerline_twist_deg is None:
            self.centerline_twist_deg = 12. * self.girths + 5

@dataclass
class JibSail:
    centerline_twist_deg: np.ndarray = None
    
    girths: np.ndarray = np.array([0.00, 1./4, 1./2, 3./4, 1.00])
    chords: np.ndarray =np.array([3.80, 2.98, 2.15, 1.33, 0.5])
    camber: np.ndarray = 0*np.array([0.01, 0.01, 0.01, 0.01, 0.01])
    camber_distance_from_luff: np.ndarray = np.array([0.5, 0.5, 0.5, 0.5, 0.5]) # starting from leading edge   
    def __post_init__(self):
        if self.centerline_twist_deg is None:
            self.centerline_twist_deg = 15. * self.girths + 7

        
# REFERENCE CSYS
# The origin of the default CSYS is located @ waterline level and aft face of the mast
# The positive x-coord: towards stern
# The positive y-coord: towards leeward side
# The positive z-coord: above the water
# To shift the default CSYS, adjust the 'reference_level_for_moments' variable.
# Shifted CSYS = original + reference_level_for_moments
# As a results the moments will be calculated around the new origin.

# yaw_reference [m] - distance from the aft of the mast towards stern, at which the yawing moment is calculated.
# sway_reference [m] - distance from the aft of the mast towards leeward side. 0 for symmetric yachts ;)
# heeling_reference [m] - distance from the water level,  at which the heeling moment is calculated.
@dataclass
class Csys:
    reference_level_for_moments: np.ndarray =  np.array([0, 0, 0])  # [yaw_reference, sway_reference, heeling_reference]


# GEOMETRY OF THE KEEL
# to estimate heeling moment from keel, does not influence the optimizer.
# reminder: the z coord shall be negative (under the water)
@dataclass
class Keel:
    center_of_lateral_resistance_upright: np.ndarray =  np.array([0, 0, -1.0])  # [m] the coordinates for a yacht standing in upright position



import numpy as np
import pandas as pd
from abc import abstractmethod
from pySailingVLM.rotations.csys_transformations import CSYS_transformations


class WindBase:
    def __init__(self, alpha_true_wind_deg, tws_ref, SOG_yacht):
        """
        Parameters
        ----------
        alpha_true_wind_deg - [deg] angle between true wind and direction of boat movement (including leeway)
        tws -  Free stream of true wind having velocity [m/s] at height z = 10 [m]
        """

        self.alpha_true_wind = np.deg2rad(alpha_true_wind_deg)  # [rad] angle between true wind and direction of boat movement (including leeway)
        self.tws_ref = tws_ref
        self.V_yacht = SOG_yacht

    @abstractmethod
    def get_true_wind_speed_at_h(self, height):
        # calc_tws_at_h = lambda h, tws_ref: tws_ref * (np.log((h) / self.roughness) / np.log(10 / self.roughness))
        # wind_speed = np.array([calc_tws_at_h(h, tws_ref) for h in heights])
        pass

    @abstractmethod
    def to_df_integral(self, csys_transformations: CSYS_transformations):
        pass

    def length_to_xyz_vector(self, wind_speed):
        tws = np.array([wind_speed * np.cos(self.alpha_true_wind),
                        wind_speed * np.sin(self.alpha_true_wind),
                        0])

        return tws

    def get_app_alfa_infs_at_h(self, tws_at_h):
        # alfa_yacht - angle between apparent wind and true wind
        # alfa_app_infs	- angle between apparent wind and direction of boat movement (including leeway)
        # (model of an 'infinite sail' is assumed == without induced wind velocity) and direction of boat movement (including leeway)

        tws_l = np.linalg.norm(tws_at_h)  # true wind speed - length of the vector
        # tws_l = np.sqrt(tws_at_h[0]*tws_at_h[0]+tws_at_h[1]*tws_at_h[1])  # true wind speed

        alfa_yacht = np.arccos((self.V_yacht * np.cos(self.alpha_true_wind) + tws_l) /
                               np.sqrt(
                                   self.V_yacht*self.V_yacht
                                   + 2*self.V_yacht*tws_l*np.cos(self.alpha_true_wind)
                                   + tws_l*tws_l))  # result in [rad]

        alpha_app_infs = self.alpha_true_wind - alfa_yacht  # result in [rad]

        return alpha_app_infs

    def get_app_infs_at_h(self, tws_at_h):
        aw_infs = np.array([tws_at_h[0] + self.V_yacht, tws_at_h[1], tws_at_h[2]])
        return aw_infs


class FlatWindProfile(WindBase):
    def __init__(self, alpha_true_wind_deg, tws_ref, SOG_yacht):
        super().__init__(alpha_true_wind_deg, tws_ref, SOG_yacht)

    def get_true_wind_speed_at_h(self, height):
        tws_at_h = self.length_to_xyz_vector(self.tws_ref)
        return tws_at_h

    def to_df_integral(self, csys_transformations: CSYS_transformations):
        obj_as_dict = {
            'Not implemented - dummy variable': 123456789,
        }
        df = pd.DataFrame.from_records(obj_as_dict, index=['Value']).transpose()
        df.reset_index(inplace=True)  # Convert the index (0-th like column) to 'regular' Column
        df = df.rename(columns={'index': 'Quantity'})
        return df


class LogWindProfile(WindBase):
    """
    roughness - Over smooth, open water, expect a value around 0.0002 m,
    while over flat, open grassland  ~ 0.03 m,
    cropland ~ 0.1-0.25 m, and brush or forest ~ 0.5-1.0 m
    """
    def __init__(self, alpha_true_wind_deg, tws_ref, SOG_yacht, roughness=0.05, reference_measurment_height=10.):
        super().__init__(alpha_true_wind_deg, tws_ref, SOG_yacht)
        self.roughness = roughness
        self.reference_measurment_height = reference_measurment_height

    def get_true_wind_speed_at_h(self, height):
        #  tws -  Free stream of true wind having velocity [m/s] at height z = 10 [m]
        wind_speed = self.tws_ref * (np.log((height) / self.roughness) / np.log(self.reference_measurment_height / self.roughness))
        tws_at_h = self.length_to_xyz_vector(wind_speed)
        return tws_at_h

    def to_df_integral(self, csys_transformations: CSYS_transformations):
        obj_as_dict = {
            'Not implemented - dummy variable': 123456789,
        }
        df = pd.DataFrame.from_records(obj_as_dict, index=['Value']).transpose()
        df.reset_index(inplace=True)  # Convert the index (0-th like column) to 'regular' Column
        df = df.rename(columns={'index': 'Quantity'})
        return df

class ExpWindProfile(WindBase):
    def __init__(self, alpha_true_wind_deg, tws_ref, SOG_yacht,
                 exp_coeff=0.1428,
                 reference_measurment_height=10.,
                 reference_water_level_for_wind_profile=0.):
        super().__init__(alpha_true_wind_deg, tws_ref, SOG_yacht)
        self.exp_coeff = exp_coeff
        self.reference_measurment_height = reference_measurment_height
        self.reference_water_level_for_wind_profile = reference_water_level_for_wind_profile
        # reference_waterline_level_for_wind_profile - this is an attempt to mimick the deck effect by lowering the sheer_above_waterline (sails' mirror)
        # while keeping the wind profile as in original geometry

    def get_true_wind_speed_at_h(self, height):
        wind_speed = self.tws_ref*pow((height - self.reference_water_level_for_wind_profile) / self.reference_measurment_height, self.exp_coeff)
        tws_at_h = self.length_to_xyz_vector(wind_speed)
        return tws_at_h

    def to_df_integral(self, csys_transformations: CSYS_transformations):
        tws_at_reference_height = self.get_true_wind_speed_at_h(self.reference_measurment_height)
        tws_length_at_reference_height = np.linalg.norm(tws_at_reference_height)

        V_app_infs_at_reference_height = self.get_app_infs_at_h(tws_at_reference_height)
        V_app_infs_length_at_reference_height = np.linalg.norm(V_app_infs_at_reference_height)
        AWA_infs_at_reference_height_deg = np.rad2deg(
            np.arctan2(V_app_infs_at_reference_height[1], V_app_infs_at_reference_height[0]))

        obj_as_dict = {
            'Reference_measurement_height': self.reference_measurment_height,
            'Reference_water_level_for_wind_profile': self.reference_water_level_for_wind_profile,
            'TWS_at_reference_height_COG.x': tws_at_reference_height[0],
            'TWS_at_reference_height_COG.y': tws_at_reference_height[1],
            'TWS_at_reference_height_COG.z': tws_at_reference_height[2],
            'TWS_length_at_reference_height': tws_length_at_reference_height,
            'V_app_infs_at_reference_height_COG.x': V_app_infs_at_reference_height[0],
            'V_app_infs_at_reference_height_COG.y': V_app_infs_at_reference_height[1],
            'V_app_infs_at_reference_height_COG.z': V_app_infs_at_reference_height[2],
            'V_app_infs_length_at_reference_height_COG': V_app_infs_length_at_reference_height,
            'AWA_infs_at_reference_height_COG_deg': AWA_infs_at_reference_height_deg,
            'AWA_infs_at_reference_height_COW_deg': AWA_infs_at_reference_height_deg - csys_transformations.leeway_deg,
        }
        df = pd.DataFrame.from_records(obj_as_dict, index=['Value']).transpose()
        df.reset_index(inplace=True)  # Convert the index (0-th like column) to 'regular' Column
        df = df.rename(columns={'index': 'Quantity'})
        return df
